"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from abc import ABC, abstractmethod
from collections import deque
from functools import cached_property
from typing import Any, Callable, Generic, Optional, TypeVar

from pydantic import BaseModel, Field, model_validator

from ..eka import Circuit
from ..interpreter import InterpretationStep

from .op_signature import ALL_EKA_OP_SIGNATURES, OpSignature, OpType


# Define generic type variables for the converter
# Each of the child classes will specify these types

# TargetType is the type returned by the convert methods, usually this will be a
# string of the program
TargetType = TypeVar("TargetType")  # pylint: disable=invalid-name

# TargetRunResultType is the type returned by running the converted circuit in the
# target language, if applicable
TargetRunResultType = TypeVar("TargetRunResultType")  # pylint: disable=invalid-name

# Define a type for the callable that maps operation and targets to instruction string
# This is usually used as type for the value in the operations_map property of each
# converter
OpAndTargetToInstrCallable = Callable[[list[str], list[str], Optional[str]], str]


class Converter(ABC, BaseModel, Generic[TargetType, TargetRunResultType]):
    """
    Abstract base class for converting Eka circuits and experiments to a specific
    format.

    This class defines the structure for converters that can transform Eka operations
    into a target language or format.
    Subclasses must implement the abstract methods to provide the actual conversion
    logic.
    And functions to evaluate the output format in terms of Eka language.

    Attributes
    ----------
    SUPPORTED_OPERATIONS : list[OpSignature]
        A set of all supported quantum operations.

    ALLOW_ERROR_MODELS: bool
        Flag indicating whether the converter allows error models to be passed during
        conversion.

    Properties
    ----------
    operations_map : dict[OpSignature, OpAndTargetToInstrCallable]
        A mapping of the supported operations to their corresponding implementation
        functions in the target language.
    eka_op_by_name : dict[str, OpSignature]
        A mapping of operation names to corresponding OpSignature objects.

    Methods
    -------
    convert(interpreted_eka: InterpretationStep) -> Any
        Convert the input interpretation step to a specific format.
    convert_circuit(input_circuit: Circuit) -> Any
        Convert a Circuit to a specific format.
    parse_target_run_outcome(run_output: TargetRunResultType)-> dict[str, list[int]]
        Parse the run output of the target language into a dictionary mapping the
        Eka channel labels to integer values measured at each shot.
    emit_circuit_program(
        input_circuit: Circuit,
        q_register: dict[str, Any],
        c_register: dict[str, Any],
    ) -> str
        Emit the full program string for a given circuit, given some quantum and
        classical register mappings.
    emit_init_instructions(
        circuit: Circuit
    ) -> tuple[str, dict[str, Any], dict[str, Any]]
        Provide the code to initialize the quantum and classical registers, and return
        the mapping from Eka channel id to register.
    emit_leaf_circuit_instruction(
        circuit: Circuit,
        quantum_channel_map: dict[str, Any],
        classical_channel_map: dict[str, Any],
    ) -> str
        Provide the code to emit an Eka instruction in the target language.

    Raises
    ------
    TypeError
        If the mapping is not a dict.
    ValueError
        If the mapping is missing any of the required keys.
    """

    model_config = {"frozen": True}  # makes instances immutable

    SUPPORTED_OPERATIONS: frozenset[OpSignature]
    ALLOW_ERROR_MODELS: bool = False

    # Separator string for else conditions in if-else constructs
    separator_for_else_in_condition: str = Field(
        default=", is_else=",
        frozen=True,
        init=False,
        description="The separator string used in the description for else conditions.",
    )

    @model_validator(mode="after")
    def _validate_ops(self):
        """
        Validate that the converter supports all the required quantum operations
        specified.

        This validator checks:

        1. operations_map builds successfully
        2. The operations_map covers all supported operations

        Returns
        -------
        self
            The validated instance

        Raises
        ------
        ValueError
            If validation fails for any of the above conditions
        """
        # 1. Check that operations_map builds successfully
        try:
            ops_map = self.operations_map
        except Exception as e:
            raise ValueError(f"Failed to build operations_map: {e}") from e

        # 2. Check that all supported operations are present in operations_map
        missing_ops = [
            op.name for op in self.SUPPORTED_OPERATIONS if op.name not in ops_map
        ]
        if missing_ops:
            raise ValueError(
                "operations_map is missing implementations for "
                f"operations: {missing_ops}"
            )

        return self

    @property
    @abstractmethod
    def operations_map(self) -> dict[str, OpAndTargetToInstrCallable]:
        """Returns a mapping of the supported operations to their corresponding
        instructions in the target language."""
        raise NotImplementedError(
            "Subclasses must implement the operations_map property."
        )

    @cached_property
    def op_by_eka_name(self) -> dict[str, OpSignature]:
        """Map of operation names to corresponding OpSignature objects."""
        return {op.name: op for op in ALL_EKA_OP_SIGNATURES}

    def convert(self, interpreted_eka: InterpretationStep) -> TargetType:
        """Convert a InterpretationStep. By default, it converts the final circuit of
        the step. This can differ from convert_circuit for converters that support
        detector and observable objects (like Stim).
        """
        return self.convert_circuit(interpreted_eka.final_circuit)

    @abstractmethod
    def convert_circuit(self, input_circuit: Circuit) -> TargetType:
        """Convert a Circuit into a program in the target language."""

    def _validate_ops_args(
        self, op_name: str, num_q_target: int, num_c_target: int
    ) -> None:
        """
        Validate the arguments for the operation.
        This validation checks that the number of quantum and classical targets
        provided for the operation matches the expected numbers defined in the
        OpSignature.

        For the BOOL_LOGIC OpType (used for condition in classical control flow),
        it allows for any positive number of classical targets greater than or equal to
        the required number, except for mono-bit boolean operations (Not and Match) that
        allows only 1 classical channel.

        The UTILS OpType operations are always valid by default.

        Parameters
        ----------
        op_name : str
            The name of the operation to validate.
        num_q_target : int
            The number of quantum targets provided for the operation.
        num_c_target : int
            The number of classical targets provided for the operation.

        Raises
        ------
        TypeError
            If num_q_target or num_c_target is not an integer.
        ValueError
            If the operation is unsupported or if the number of targets does not match.
        """
        if not isinstance(num_q_target, int):
            raise TypeError(f"{op_name} quantum target must be an integer")
        if not isinstance(num_c_target, int):
            raise TypeError(f"{op_name} classical target must be an integer")

        if op_name not in self.op_by_eka_name:
            raise ValueError(f"Unsupported operation '{op_name}'")

        eka_sig = self.op_by_eka_name[op_name]
        if eka_sig.quantum_input != num_q_target:
            raise ValueError(
                f"{op_name} quantum target must have "
                f"{eka_sig.quantum_input} qubits, "
                f"but got {num_q_target}."
            )
        if eka_sig.classical_input != num_c_target:
            if eka_sig.op_type == OpType.UTILS:
                # pass utils operations, used for comments, etc.
                return
            if eka_sig.op_type == OpType.BOOL_LOGIC:
                if (
                    num_c_target <= eka_sig.classical_input
                    and eka_sig.classical_input != 1
                ):
                    # Special case: allow n classical bits for boolean logic ops as long
                    # as n >= required bits (and need at least 2 bits)
                    # And this doesn't apply to mono-bit boolean ops (Not and Match)
                    raise ValueError(
                        f"{op_name} must have at least {eka_sig.classical_input}"
                        " classical bits"
                    )
                return
            raise ValueError(
                f"{op_name} classical target must have "
                f"{eka_sig.classical_input} bits, "
                f"but got {num_c_target}."
            )

    @staticmethod
    def _validate_import_prefix(import_prefix: str) -> None:
        """
        Validate the import prefix string.
        It must be a string that ends with a dot (.)
        or be an empty string.

        Parameters
        ----------
        import_prefix : str
            The import prefix string to validate.

        Raises
        ------
        TypeError
            If import_prefix is not a string.
        ValueError
            If import_prefix is not empty and does not end with a dot (.).
        """
        if not isinstance(import_prefix, str):
            raise TypeError("import_prefix must be a string")
        if import_prefix == "":
            return
        if not import_prefix.endswith("."):
            raise ValueError(f"import_prefix '{import_prefix}' must end with a dot (.)")

    @staticmethod
    @abstractmethod
    def parse_target_run_outcome(
        run_output: TargetRunResultType,
    ) -> dict[str, int | list[int]]:
        """Parse the run output of the target language into a dictionary mapping the
        eka channel labels to values measured at each shot."""

    # pylint: disable=too-many-branches
    def emit_circuit_program(
        self,
        input_circuit: Circuit,
        q_register: dict[str, str],
        c_register: dict[str, str | None],
    ) -> str:
        """Emit the full program string for a given circuit, given some quantum and
        classical register mappings.

        This method process the circuit in a depth-first manner, handling nested
        circuits and control flow constructs (if-else) appropriately. It uses a stack
        to keep track of circuits to process, and maintains an indentation level for
        formatting the output.

        Parameters
        ----------
        input_circuit : Circuit
            The input circuit to emit the program for.
        q_register : dict[str, str]
            The mapping from eka quantum channel ids to target quantum register names.
        c_register : dict[str, str | None]
            The mapping from eka classical channel ids to target classical register
            names.

        Returns
        -------
        str
            The emitted program as a string.
        """
        if not isinstance(input_circuit, Circuit):
            raise TypeError("Input must be a Circuit instance.")

        program_lines = []

        stack_to_process = deque([input_circuit])
        indent_level = 0
        indent = "    "

        while stack_to_process:
            processed_circuit = stack_to_process.popleft()
            if processed_circuit.name == "empty_circuit":
                continue

            if processed_circuit.name in self.op_by_eka_name:
                processed_op = self.op_by_eka_name[processed_circuit.name]

                if processed_op.op_type == OpType.UTILS:
                    match processed_op.name:
                        case "indent_more":
                            indent_level += 1
                        case "indent_less":
                            indent_level = max(0, indent_level - 1)
                        case "comment":
                            comment_str = self.emit_leaf_circuit_instruction(
                                processed_circuit, {}, {}
                            )
                            program_lines.append(
                                f"{indent * indent_level}{comment_str}"
                            )
                else:
                    # Leaf operation: just append op strings
                    op_str = self.emit_leaf_circuit_instruction(
                        processed_circuit, q_register, c_register
                    )
                    for line in op_str.splitlines():
                        program_lines.append(f"{indent * indent_level}{line}")

            elif hasattr(processed_circuit, "_loom_ifelse_marker"):
                condition_instruction = self.emit_leaf_circuit_instruction(
                    processed_circuit.condition_circuit,
                    quantum_channel_map=q_register,
                    classical_channel_map=c_register,
                )
                is_else_present = False
                if_else_instruction = [
                    (Circuit(name="classical_if", description=condition_instruction)),
                    (Circuit(name="indent_more")),
                    (processed_circuit.if_circuit),
                ]
                if processed_circuit.else_circuit.name != "empty_branch":
                    is_else_present = True
                    if_else_instruction += [
                        (Circuit(name="indent_less")),
                        (
                            Circuit(
                                name="classical_else", description=condition_instruction
                            )
                        ),
                        (Circuit(name="indent_more")),
                        (processed_circuit.else_circuit),
                    ]
                if_else_instruction += [
                    (Circuit(name="indent_less")),
                    (
                        Circuit(
                            name="end_if",
                            description=condition_instruction
                            + self.separator_for_else_in_condition
                            + str(is_else_present),
                        )
                    ),
                ]
                for line in reversed(if_else_instruction):
                    stack_to_process.appendleft(line)

            else:  # Nested circuit
                nested_instructions = [
                    c for tick in processed_circuit.circuit for c in tick
                ]
                if not processed_circuit.name.startswith("wrapped_"):
                    start_comment = [
                        Circuit(
                            name="comment",
                            description=f"Start of circuit: {processed_circuit.name}",
                        )
                    ]
                    nested_instructions = start_comment + nested_instructions
                    nested_instructions += [
                        Circuit(
                            name="comment",
                            description=f" End of circuit: {processed_circuit.name}",
                        )
                    ]
                for line in reversed(nested_instructions):
                    stack_to_process.appendleft(line)

        return "\n".join(program_lines)

    @abstractmethod
    def emit_init_instructions(
        self, circuit: Circuit
    ) -> tuple[str, dict[str, Any], dict[str, Any]]:
        """Provide the python code (as a string) to initializes the
        quantum and classical registers, and return the mapping from eka channel id to
        register."""

    @abstractmethod
    def emit_leaf_circuit_instruction(
        self,
        circuit: Circuit,
        quantum_channel_map: dict[str, Any],
        classical_channel_map: dict[str, Any],
    ) -> str:
        """Provide the python code (as a string) to emit an Eka instruction in the
        target language."""



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from loom.eka import Circuit
from loom.eka.operations import (
    MeasureLogicalX,
    MeasureLogicalY,
    MeasureLogicalZ,
    LogicalMeasurement,
)
from loom.eka.stabilizer import Stabilizer

from loom.interpreter.syndrome import Syndrome
from .generate_syndromes import generate_syndromes
from .generate_detectors import generate_detectors
from ..interpretation_step import InterpretationStep
from ..logical_observable import LogicalObservable


def measurelogicalpauli(
    interpretation_step: InterpretationStep,
    operation: MeasureLogicalX | MeasureLogicalY | MeasureLogicalZ,
    same_timeslice: bool,
    debug_mode: bool,  # pylint: disable=unused-argument
) -> InterpretationStep:
    """
    Measure a logical Pauli operator. Y logical measurements are not supported yet.

    The algorithm is the following:
    
    - A.) Measure all data qubits in the block
    - B.) Create and add single-qubit stabilizers to `measured_single_qubit_stabilizers`
    - C.) Update Syndromes for all stabilizers involved in the data qubits measured
    - D.) Create the logical observable including measured data qubits and all \
    previous corrections

    Parameters
    ----------
    interpretation_step : InterpretationStep
        The interpretation step to which the operation should be applied.
    operation : MeasureLogicalX | MeasureLogicalY | MeasureLogicalZ
        The operation to be applied, can either be a logical X, Y or Z measurement.
    same_timeslice : bool
        Flag indicating whether the operation is part of the same timestep as the
        previous operation.
    debug_mode : bool
        Flag indicating whether the interpretation should be done in debug mode.
        Activating debug mode will enable commutation validation for Block.

    Returns
    -------
    InterpretationStep
        New InterpretationStep containing all modifications due to the logical pauli
        measurement.
    """
    block = interpretation_step.get_block(operation.input_block_name)
    logical_qubit_index = operation.logical_qubit

    # Check for correct operation
    match operation.__class__.__name__:
        case "MeasureLogicalX":
            basis = "X"
        case "MeasureLogicalY":
            raise ValueError("Logical measurement in Y basis is not supported")
        case "MeasureLogicalZ":
            basis = "Z"
        case _:
            raise ValueError(f"Operation {operation.__class__.__name__} not supported")

    # A) - Measure all data qubits in a block and keep track of the Cbits
    meas_circuit_seq = []

    # Add Hadamard layer for X basis
    if basis == "X":
        hadamard_layer = [
            Circuit(
                "H", channels=interpretation_step.get_channel_MUT(str(q), "quantum")
            )
            for q in block.data_qubits
        ]
        meas_circuit_seq += [hadamard_layer]

    # Create the measurement layer in Z basis
    cbit_labels = [f"c_{qubit}" for qubit in block.data_qubits]
    measurements = tuple(
        interpretation_step.get_new_cbit_MUT(label) for label in cbit_labels
    )
    measurement_layer = [
        Circuit(
            "measurement",
            channels=[
                interpretation_step.get_channel_MUT(
                    str(qubit), "quantum"
                ),  # qubit to measure
                interpretation_step.get_channel_MUT(
                    f"{cbit[0]}_{cbit[1]}", "classical"
                ),  # associated classical bit
            ],
        )
        for qubit, cbit in zip(block.data_qubits, measurements, strict=True)
    ]

    meas_circuit_seq += [measurement_layer]
    meas_circuit = Circuit(
        f"Measure logical {basis} of {block.unique_label}", circuit=meas_circuit_seq
    )

    # Append the circuit
    interpretation_step.append_circuit_MUT(meas_circuit, same_timeslice)

    # B) - Create and add single-qubit stabilizers
    # to `measured_single_qubit_stabilizers`
    measured_single_qubit_stabilizers = {
        Stabilizer(
            pauli=basis,
            data_qubits=(q,),
        )
        for q in block.data_qubits
    }
    interpretation_step.update_measured_single_qubit_stabilizers_MUT(
        block_id=block.uuid,
        new_single_qubit_stabilizers=measured_single_qubit_stabilizers,
    )

    # C) - Update Syndromes for all stabilizers involved in the data qubits measured
    # pylint: disable-next=unused-variable
    meaured_single_qubit_syndromes = (
        Syndrome(
            stabilizer=stab.uuid,
            measurements=tuple(
                cbit
                for cbit in measurements
                if cbit[0].split("_")[1] == str(stab.data_qubits[0])
            ),
            block=block.uuid,
            round=-1,  # should not be associated with any round
            labels={stab.uuid: stab.data_qubits[0]},
        )
        # only use the stabilizers of the right pauli type
        for stab in measured_single_qubit_stabilizers
        if stab.pauli == basis
    )

    # Only use the stabilizers of the right pauli type
    relevant_stabs = [stab for stab in block.stabilizers if set(stab.pauli) == {basis}]
    # Get the classical bits associated with these stabilizers
    stab_cbits = [
        [
            cbit
            for cbit in measurements
            if cbit[0].split("_")[1] in map(str, stab.data_qubits)
        ]
        for stab in relevant_stabs
    ]
    # Create new Syndromes from these measurements
    new_syndromes = generate_syndromes(
        interpretation_step, relevant_stabs, block, stab_cbits
    )
    # Create Detectors for the new syndromes
    new_detectors = generate_detectors(interpretation_step, new_syndromes)

    # D) - Create the logical observable including measured
    # data qubits and all previous corrections
    if basis == "X":
        logical_qubit = block.logical_x_operators[logical_qubit_index]
        operator_updates = interpretation_step.logical_x_operator_updates
    else:
        logical_qubit = block.logical_z_operators[logical_qubit_index]
        operator_updates = interpretation_step.logical_z_operator_updates

    corrections = list(operator_updates.get(logical_qubit.uuid, ()))
    qubits_in_logical = [str(q) for q in logical_qubit.data_qubits]
    logical_measurements = [
        cbit for cbit in measurements if cbit[0].split("_")[1] in qubits_in_logical
    ]
    logical_observable = LogicalObservable(
        label=f"{block.unique_label}_{basis}_{logical_qubit_index}",
        measurements=logical_measurements + corrections,
    )
    interpretation_step.logical_measurements[
        LogicalMeasurement(blocks=(block.unique_label,), observable=basis)
    ] = tuple(logical_measurements + corrections)
    interpretation_step.append_detectors_MUT(new_detectors)
    interpretation_step.append_syndromes_MUT(new_syndromes)
    interpretation_step.logical_observables += (logical_observable,)

    return interpretation_step



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from __future__ import annotations

from .circuit import Channel, Circuit
from .stabilizer import Stabilizer
from .syndrome_circuit import SyndromeCircuit
from .tanner_graphs import TannerGraph
from .utilities.graph_matrix_utils import (
    minimum_edge_coloring,
    extract_subgraphs_from_edge_labels,
)


def coloration_circuit(
    t_graph: TannerGraph,
) -> dict[tuple[int, ...], list[tuple[int, ...]]]:
    """
    Computes the entangling layer order for a CSS code specified through a Tanner graph.
    This layering is obtained by the coloration circuit algorithm described in
    arXiv:2109.14609. In summary, the algorithm measures X and Z stabilizers in
    sequence, but optimizes the circuit depth for each type. This is done by extracting
    the X and Z subcomponents of the Tanner graph, and by obtaining a minimum edge
    coloring for each one. The sequence of entangling gates then corresponds to all the
    edges in each coloring layer, and then switching to the stabilizer type once the
    first component has been exhausted. Since the Tanner Graph is bipartite, this
    approach results in an upper bound of of max(deg(t_graph_x)) + max(deg(t_graph_z))
    for the total number of entangling layers, where t_graph_x and t_graph_z are the X
    and Z tanner subcomponents, respectively.

    The function returns a dictionary, where the keys correspond to each check nodes
    associated with a Stabilizer, and the values are the lists of data qubit channels
    in the precise order in which the entangling gates should be applied. Idling steps
    are added for Z(X) checks when the entangling gates between X(Z) ancillas and datas
    take place.

    Parameters
    ----------
    t_graph : TannerGraph
        Tanner graph associated with a CSS code.

    Returns
    -------
    check_to_data: dict[tuple[int,...],list[tuple[int,...]]]
        Dictionary mapping check nodes to lists of data nodes, properly ordered for
        entangling operations.
    """

    # Initialize list containing CNOT actions
    check_to_data = {node: [] for node in t_graph.x_nodes} | {
        node: [] for node in t_graph.z_nodes
    }

    # Extract subgraphs according to cardinality
    t_graph_x, t_graph_z = t_graph.get_components()

    # Loop over both subcomponents
    for t_card, t_card_opp in [(t_graph_x, t_graph_z), (t_graph_z, t_graph_x)]:

        # Obtain minimum edge coloring for the subgraph
        coloring = minimum_edge_coloring(t_card.graph)

        # Apply entangling gates for each color layer
        for _, colored_edges in coloring.items():

            # Store colored checks to identify idling ones later
            colored_checks = set()

            # Simultaneously apply all gates supported edges of t_card with same color
            for edge in colored_edges:
                if t_card.graph.nodes[edge[1]]["label"] == "data":
                    check_node, data_node = edge
                else:
                    check_node, data_node = edge[::-1]
                check_to_data[check_node].append(data_node)
                colored_checks.add(check_node)

            # Add idling step for qubits not participating in the layer
            # These include a subset of the local check type and those of opposite type
            idle_nodes = list(t_card_opp.check_nodes) + [
                n for n in t_card.check_nodes if n not in colored_checks
            ]
            for node in idle_nodes:
                check_to_data[node].append(())

    return check_to_data


def cardinal_circuit(
    t_graph: TannerGraph,
) -> dict[tuple[int, int], list[tuple[int, int]]]:
    """
    Computes the entangling layer order for a HGP code specified through a Tanner graph
    with cardinal edge labelling. This layering is obtained by the cardinal circuit
    algorithm described in arXiv:2109.14609. Importantly, the ordering over which the
    cardinalities are examined is CRUCIAL for commutation relations. Here it has been
    fixed to ["E", "N", "S", "W"]. Only a subset of permutations from this order will
    ensure proper commutation of stabilizers.

    The function returns a dictionary, where the keys correspond to each check nodes
    associated with a Stabilizer, and the values are the lists of data qubit channels
    in the precise order in which the entangling gates should be applied. Idling steps
    are added when ancillas do not participate in a given time step.

    Parameters
    ----------
    t_graph : TannerGraph
        Tanner graph associated with the corresponding HGP code.

    Returns
    -------
    check_to_data: dict[tuple[int,int],list[tuple[int,int]]]
        Dictionary mapping check nodes to lists of data nodes, properly ordered for
        entangling operations.
    """

    # Check that input Tanner Graph contains proper edge labels
    attribute_names = {
        key for _, _, attr in t_graph.graph.edges(data=True) for key in attr
    }
    if "cardinality" not in attribute_names:
        raise ValueError("All edges most have a 'cardinality' attribute.")

    edge_labels = {label for _, _, label in t_graph.graph.edges(data="cardinality")}
    if not all(label in ["E", "N", "S", "W"] for label in edge_labels):
        raise ValueError("Cardinality labels should be either 'E', 'W', 'N' or 'S'.")

    # Initialize list containing CNOT actions
    check_to_data = {node: [] for node in t_graph.x_nodes} | {
        node: [] for node in t_graph.z_nodes
    }

    # Extract subgraphs according to cardinality
    t_sub_dict = extract_subgraphs_from_edge_labels(t_graph.graph, "cardinality")

    # Loop over cardinalities in specific valid order
    for cardinality in ["E", "N", "S", "W"]:

        # Extract subgraph associated with cardinality
        t_card = t_sub_dict[cardinality]

        # Obtain minimum edge coloring for the cardinal subgraph
        coloring = minimum_edge_coloring(t_card)

        # Apply entangling gates for each color layer
        for _, colored_edges in coloring.items():

            # Simultaneously apply all gates supported edges of t_card with same color
            for edge in colored_edges:
                if t_card.nodes[edge[1]]["label"] == "data":
                    check_node, data_node = edge
                else:
                    check_node, data_node = edge[::-1]
                check_to_data[check_node].append(data_node)

            # Add idling step for qubits not participating in the layer
            active_nodes = set(node for edge in colored_edges for node in edge)
            idle_nodes = set(t_graph.x_nodes + t_graph.z_nodes) - active_nodes
            for node in idle_nodes:
                check_to_data[node].append(())

    return check_to_data


def generate_stabilizer_and_syndrome_circuits_from_algorithm(
    t_graph: TannerGraph,
    algorithm: str,
) -> tuple[list[Stabilizer], list[SyndromeCircuit]]:
    """
    Generates stabilizers and syndrome circuits for a Tanner graph t_graph using the
    a supported circuit algorithm such as the cardinal or the coloration algorithms.

    Parameters
    ----------
    t_graph : TannerGraph
        Tanner graph associated with the code.

    algorithm : str
        Algorithm to be used to find the entangling schedule.

    Returns
    -------
    tuple[list[Stabilizer], list[SyndromeCircuit]]
        A list of Stabilizer objects describing the code and alist of associated
        SyndromeCircuit objects.
    """

    match algorithm:
        case "cardinal":
            entangling_schedule = cardinal_circuit(t_graph)
        case "coloration":
            entangling_schedule = coloration_circuit(t_graph)
        case _:
            raise NotImplementedError(
                f"The scheduling algorithm {algorithm} is not implemented."
                " Must be 'cardinal' or 'coloration."
            )

    # Leverage a circuit algorithm schedule to generate stabilizers and circuits
    syndrome_circuits = []
    stabilizers = []

    # Loop over check nodes of the Tanner graph
    for node in t_graph.x_nodes + t_graph.z_nodes:
        pauli = t_graph.graph.nodes[node]["label"] * t_graph.graph.degree(node)

        # Compute stabilizer associated with check node
        stabilizer = Stabilizer(
            pauli=pauli,
            data_qubits=[q for q in entangling_schedule[node] if q],
            ancilla_qubits=[node],
        )
        stabilizers.append(stabilizer)

        # Compute syndrome circuit associated with check node
        syndrome_circuit = extract_syndrome_circuit(pauli, entangling_schedule[node])
        syndrome_circuits.append(syndrome_circuit)

    return stabilizers, syndrome_circuits


def extract_syndrome_circuit(
    pauli: str, entangling_order: list[tuple[int, ...]]
) -> SyndromeCircuit:
    """
    Generate the syndrome extraction circuit for a given stabilizer.

    Parameters
    ----------
    pauli : str
        The Pauli string defining the stabilizer.
    entangling_order : list[tuple[int,...]]
        The order in which data qubits are entangled with the ancilla qubit.

    Returns
    -------
    SyndromeCircuit
        The syndrome extraction circuit for the given stabilizer.
    """
    # Extract parameters
    pauli_type = next(iter(pauli))
    name = f"{pauli}_syndrome_extraction"
    weight = len(pauli)

    # Define channels
    data_channels = [Channel(type="quantum", label=f"d{i}") for i in range(weight)]
    cbit_channel = Channel(type="classical", label="c0")
    ancilla_channel = Channel(type="quantum", label="a0")

    # Define Hadamard gates
    hadamard1 = [Circuit("H", channels=[ancilla_channel])]
    hadamard2 = [Circuit("H", channels=[ancilla_channel])]

    # Define mapping between gate action and data qubit inside stabilizer

    entangle_ancilla = []
    q_counter = 0

    for qubit in entangling_order:
        if qubit:
            engtangling_gate = Circuit(
                f"C{pauli_type}",
                channels=[ancilla_channel, data_channels[q_counter]],
            )
            entangle_ancilla.append([engtangling_gate])
            q_counter += 1
        else:
            entangle_ancilla.append([])

    # Add ancilla measurement and reset
    measurement = [Circuit("Measurement", channels=[ancilla_channel, cbit_channel])]
    reset = [Circuit("Reset_0", channels=[ancilla_channel])]

    # Compose circuit operations as a list
    circuit_list = [reset, hadamard1] + entangle_ancilla + [hadamard2, measurement]

    # Build the syndrome extraction circuit
    syndrome_circuit = SyndromeCircuit(
        pauli=pauli_type * weight,
        name=name,
        circuit=Circuit(
            name=name,
            circuit=circuit_list,
            channels=data_channels + [ancilla_channel, cbit_channel],
        ),
    )

    return syndrome_circuit



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from loom.eka import Circuit, Stabilizer, PauliOperator
from loom.eka.utilities import DiagonalDirection
from loom.interpreter import InterpretationStep

from .y_wall_out_utilities import (
    generate_teleportation_finalization_circuit_with_updates,
    qubit_mapper,
    get_final_block_syndrome_measurement_cnots_circuit,
    move_logicals_and_append_evolution_and_updates,
)
from ..move_block import (
    find_swap_then_qec_qubit_initializations,
    generate_syndrome_measurement_circuit_and_cbits,
    generate_and_append_block_syndromes_and_detectors,
)

from ..utilities import (
    generate_syndrome_extraction_circuits,
    find_relative_diagonal_direction,
)
from ...code_factory import RotatedSurfaceCode


def y_wall_out_final_swap_then_qec(
    interpretation_step: InterpretationStep,
    block: RotatedSurfaceCode,
    idle_side_directions: DiagonalDirection,
    same_timeslice: bool,
    debug_mode: bool,
) -> InterpretationStep:
    """
    Apply the final swap-then-QEC operation for Y-wall-out recombination. By the end
    of this operation, the block has been aligned with its original position before the
    y_wall_out operation in a fault-tolerant manner.

    The operation consists of the following steps:

    A.) Stabilizer evolution and updates

        - A.1) Move stabilizers and define evolution in the interpretation step
        - A.2) Update stabilizer updates in interpretation step

    B.) Logical operators evolution and updates

    C.) Generate final Block and appropriate SyndromeCircuits

        C.1) Create a mock block to find the new syndrome circuits
        C.2) Find the starting qubit direction
        C.3) Find syndrome circuits using the mock block
        C.4) Create the new block with the new syndrome circuits

    D.) Circuit Generation

        D.1) Find the qubits to initialize (data and ancilla) and teleportation pairs
        D.2) Generate the initialization circuit
        D.3) Generate the CNOTs circuit for syndrome measurement
        D.4) Generate the teleportation finalization circuit
        D.5) Generate the syndrome measurement circuit and classical bits
        D.6) Append the full circuit to the interpretation step

    E.) Update Block history and evolution

    F.) Generate and append Block syndromes and detectors


    Parameters
    ----------
    interpretation_step: InterpretationStep
        The interpretation step.
    block: RotatedSurfaceCode
        The block to which the operation will be applied.
    idle_side_directions: DiagonalDirection
        The direction towards the idle side of the block was moved during the
        first swap-then-QEC operation.
    same_timeslice: bool
        Whether to append the generated circuit to the same timeslice as the last
        circuit in the interpretation step.
    debug_mode: bool
        Whether to enable debug mode.
    """

    final_moving_directions = idle_side_directions.opposite()

    # A.) Stabilizer evolution and updates
    new_stabilizers = move_stabilizers_and_append_evolution_and_updates(
        interpretation_step,
        block.stabilizers,
        final_moving_directions,
    )

    # B.) Logical operators evolution and updates
    new_logical_x, new_logical_z = move_logicals_and_append_evolution_and_updates(
        interpretation_step, block, final_moving_directions
    )

    # C.) Generate final Block and appropriate SyndromeCircuits
    final_block = generate_final_block_with_syndrome_circuits(
        block.unique_label, new_stabilizers, new_logical_x, new_logical_z, debug_mode
    )

    # D.) Circuit Generation
    # D.1) Find the qubits to initialize (data and ancilla) and teleportation pairs
    (
        anc_qubits_to_init_final,
        data_qubits_to_init_final,
        teleportation_qubit_pairs_final,
    ) = find_swap_then_qec_qubit_initializations(
        block.stabilizers, final_moving_directions
    )
    # D.2) Generate the initialization circuit
    second_swap_then_qec_reset_circuit = Circuit(
        name=("Initialization of qubits for second swap-then-qec"),
        circuit=[
            [
                Circuit(
                    f"reset_{'0' if pauli == 'Z' else '+'}",
                    channels=[interpretation_step.get_channel_MUT(q)],
                )
                for pauli in ["X", "Z"]
                for q in data_qubits_to_init_final[pauli]
                + anc_qubits_to_init_final[pauli]
            ]
        ],
    )

    # D.3) Generate the CNOTs circuit for syndrome measurement
    second_swap_then_qec_cnots_circuit = (
        get_final_block_syndrome_measurement_cnots_circuit(
            final_block,
            interpretation_step,
            True,
            idle_side_directions,
            anc_qubits_to_init_final,
        )
    )

    # D.4) Generate the teleportation finalization circuit
    second_swap_then_qec_teleportation_finalization_circuit = (
        generate_teleportation_finalization_circuit_with_updates(
            interpretation_step,
            final_block,
            anc_qubits_to_init_final,
            teleportation_qubit_pairs_final,
        )
    )
    # D.5) Generate the syndrome measurement circuit and classical bits
    second_swap_then_qec_measurement_circuit, second_swap_then_qec_cbits = (
        generate_syndrome_measurement_circuit_and_cbits(
            interpretation_step,
            final_block,
        )
    )

    # D.6) Append the full circuit to the interpretation step
    interpretation_step.append_circuit_MUT(
        Circuit(
            name="Second swap-then-QEC for Y-wall-out recombination",
            circuit=Circuit.construct_padded_circuit_time_sequence(
                (
                    (second_swap_then_qec_reset_circuit,),
                    (second_swap_then_qec_cnots_circuit,),
                    (
                        second_swap_then_qec_measurement_circuit,
                        second_swap_then_qec_teleportation_finalization_circuit,
                    ),
                )
            ),
        ),
        same_timeslice=same_timeslice,
    )

    # E.) Update Block history and evolution
    interpretation_step.update_block_history_and_evolution_MUT((final_block,), (block,))

    # F.) Generate and append Block syndromes and detectors
    generate_and_append_block_syndromes_and_detectors(
        interpretation_step, final_block, second_swap_then_qec_cbits
    )

    return interpretation_step


def generate_final_block_with_syndrome_circuits(
    block_unique_label: str,
    new_stabilizers: list[Stabilizer],
    new_logical_x: PauliOperator,
    new_logical_z: PauliOperator,
    debug_mode: bool,
) -> RotatedSurfaceCode:
    """
    Generate the final block after relocating the recombined block to its original
    position, with the appropriate syndrome extraction circuits.

    Parameters
    ----------
    block_unique_label: str
        The unique label of the block.
    new_stabilizers: list[Stabilizer]
        The stabilizers of the new block.
    new_logical_x: PauliOperator
        The logical X operator of the new block.
    new_logical_z: PauliOperator
        The logical Z operator of the new block.
    debug_mode: bool
        Whether to enable debug mode.

    Returns
    -------
    RotatedSurfaceCode
        The new block with the appropriate syndrome extraction circuits.
    """

    # C.1) Create a mock block to find the new syndrome circuits
    mock_block = RotatedSurfaceCode(
        unique_label=block_unique_label,
        stabilizers=new_stabilizers,
        logical_x_operators=(new_logical_x,),
        logical_z_operators=(new_logical_z,),
    )
    config, pivot_corners = mock_block.config_and_pivot_corners
    if config != 3:
        raise ValueError(
            f"New block must have type-3 corner configuration, found type-{config}."
        )

    # C.2) Find the starting qubit direction
    # It should be the opposite direction of the short_end_corner
    short_end_corner = pivot_corners[3]
    top_left_corner = mock_block.upper_left_qubit
    starting_diag_direction = find_relative_diagonal_direction(
        top_left_corner, short_end_corner
    )

    # C.3) Find syndrome circuits using the mock block
    new_synd_circ_tuple, new_stabilizer_to_circuit = (
        generate_syndrome_extraction_circuits(mock_block, starting_diag_direction)
    )

    # C.4) Create the new block with the new syndrome circuits
    new_block = RotatedSurfaceCode(
        stabilizers=mock_block.stabilizers,
        logical_x_operators=mock_block.logical_x_operators,
        logical_z_operators=mock_block.logical_z_operators,
        syndrome_circuits=new_synd_circ_tuple,
        stabilizer_to_circuit=new_stabilizer_to_circuit,
        unique_label=block_unique_label,
        skip_validation=not debug_mode,
    )

    return new_block


def move_stabilizers_and_append_evolution_and_updates(
    interpretation_step: InterpretationStep,
    stabilizers: list[Stabilizer],
    move_direction: DiagonalDirection,
) -> list[Stabilizer]:
    """
    Move stabilizers in the given direction and append their evolution and updates
    to the interpretation step.

    Parameters
    ----------
    interpretation_step: InterpretationStep
        The interpretation step to update.
    stabilizers: list[Stabilizer]
        The stabilizers to move.
    move_direction: DiagonalDirection
        The direction to move the stabilizers.

    Returns
    -------
    list[Stabilizer]
        The new stabilizers after the move.
    """
    # A.1) Move stabilizers and define evolution in the interpretation step
    new_stabilizers = []
    stab_evolution = {}
    for stab in stabilizers:
        new_stab = Stabilizer(
            stab.pauli,
            [qubit_mapper(dq, move_direction) for dq in stab.data_qubits],
            ancilla_qubits=[
                qubit_mapper(aq, move_direction) for aq in stab.ancilla_qubits
            ],
        )
        new_stabilizers.append(new_stab)
        stab_evolution[new_stab.uuid] = (stab.uuid,)

    # A.2) Update stabilizer updates in interpretation step
    for new_stab_uuid, old_stabs_uuid in stab_evolution.items():
        # Only one parent per new stabilizer
        old_stab_uuid = old_stabs_uuid[0]
        if old_stab_uuid in interpretation_step.stabilizer_updates:
            interpretation_step.stabilizer_updates[new_stab_uuid] = (
                interpretation_step.stabilizer_updates.pop(old_stab_uuid)
            )

    interpretation_step.stabilizer_evolution.update(stab_evolution)
    return new_stabilizers



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from collections import defaultdict
from enum import Enum
import math
from typing import Any, Callable, Optional, Protocol, runtime_checkable

from pydantic import BaseModel, Field, field_validator, model_validator, PrivateAttr

from ..eka import Circuit


# ====== Type / Aliases =====


# Callable type for error probability, return a list of float given a gate name and an
# optional time.
# Allow returning a list of floats for error probabilities. Some noise instructions
# may require multiple parameters (e.g., pauli_channel).
@runtime_checkable
# pylint: disable=too-few-public-methods
class ErrorProbProtocol(Protocol):
    """Define a protocol for error probability callables."""

    def __call__(
        self, time_from_start: Optional[float], time_of_tick: Optional[float]
    ) -> list[float]:
        """Compute error probabilities for a given gate and time parameters.

        Parameters
        ----------
        time_from_start : Optional[float]
            Cumulative time from the start of the circuit.
        time_of_tick : Optional[float]
            Duration or idle time within the current tick.

        Returns
        -------
        list[float]
            List of error probabilities.
        """
        ...  # pylint: disable=unnecessary-ellipsis


GateErrorProbProtocol = Callable[[Optional[float]], list[float]]


class ApplicationMode(str, Enum):
    """
    Enum-like class to define the application mode of the error model.
    """

    BEFORE_GATE = "before_gate"
    AFTER_GATE = "after_gate"
    END_OF_TICK = "end_of_tick"
    IDLE_END_OF_TICK = "idle_end_of_tick"


# Error model shall define the types of errors that can occur in a quantum circuit,
# In order to properly map to noise instructions in the different backends.
class ErrorType(Enum):
    """Provides a set of error types that can be used to define the error model for a
    quantum circuit.

    Each error type has a label and a number of parameters that it expects.
    Also provides a method to validate the parameters for the error type by checking
    they are a proper probability distribution (sum doesn't exceed 1 if multiple
    parameters) and that they are in the range [0, 1].
    """

    PAULI_X = ("pauli_x", 1)
    PAULI_Y = ("pauli_y", 1)
    PAULI_Z = ("pauli_z", 1)
    PAULI_CHANNEL = ("pauli_channel", 3)
    BIT_FLIP = ("bit-flip", 1)
    PHASE_FLIP = ("phase-flip", 1)
    DEPOLARIZING1 = ("depolarizing1", 1)
    DEPOLARIZING2 = ("depolarizing2", 1)

    def __init__(self, label: str, param_count: int):
        self.label = label
        self.param_count = param_count

    def validate_params(self, params: list[float]) -> None:
        """Raise ValueError if params are not valid for this error type."""
        if len(params) != self.param_count:
            raise ValueError(
                f"{self.name} expects {self.param_count} parameter(s), "
                f"got {len(params)}."
            )

        if not all(isinstance(p, (float, int)) for p in params):
            raise TypeError(f"All parameters for {self.name} must be numbers.")

        if not all(0 <= p <= 1 for p in params):
            raise ValueError(f"All parameters for {self.name} must be in [0, 1].")

        if self.param_count > 1:
            if sum(params) > 1.0:
                raise ValueError(f"{self.name} parameters sum must not exceed 1.0.")


# ====== CircuitErrorModel Class =====
class CircuitErrorModel(BaseModel):
    """
    Define a circuit error model that can be used to simulate errors in quantum
    circuits.
    This model can be time-dependent or not, having error applied in different modes,
    and can define error probabilities for each gate in the circuit or for each tick in
    the circuit.
    It is designed to be used with the Circuit class and its operations.

    This class is very general and allows for very flexible error modeling, however, it
    is a bit tedious to work with, so we recommend to define subclasses that
    define specific error models for your use case.

    Parameters
    ----------
    circuit : Circuit
        The quantum circuit to which the error model will be applied.
        This is frozen after initialization, so it cannot be changed.
    error_type : ErrorType
        The type of error that the model will apply to the circuit.
        This is frozen after initialization, so it cannot be changed.
    is_time_dependent : bool
        Whether the error model is time-dependent or not.
        If True, the model will use gate_durations to compute error probabilities.
        If False, the model will not use gate_durations.
        This is frozen after initialization, so it cannot be changed.
    application_mode : ApplicationMode
        The mode in which the error is applied to the circuit.
        It can be BEFORE_GATE, AFTER_GATE, or END_OF_TICK.
        This is frozen after initialization, so it cannot be changed.
    gate_durations : Optional[dict[str, float]]
        A dictionary mapping gate names to their execution times.
        This is only used if the model is time-dependent.
        If the model is not time-dependent, this can be None.
        If provided, it  must assign a duration to each gate type present in the
        circuit.
    gate_error_probabilities : Optional[dict[str, GateErrorProbProtocol]]
        A dictionary mapping gate names to a callable that returns the error probability
        for that gate.
        If the model is time-dependent, the callable can take an optional time
        parameter.
        If the model is not time-dependent, the callable should not take any
        parameters.
        If one gate isn't in the dictionary, it will default to a callable that
        returns 0.0.
    global_time_error_probability : Callable[[Optional[float]], list[float]]
        A callable that returns the error probability at a specific time in the circuit.
        It can take an optional time parameter, which represents some information in the
        current tick:

        - If the application mode is END_OF_TICK, it represents the duration of the tick
        - If the application mode is IDLE_END_OF_TICK, it represents the idle times
            of the channel in the tick.

        The function must be well-defined at t = 0 (for both inputs).

    """

    # Pydantic configuration, this is necessary to allow arbitrary types in the model.
    # In particular, we need to allow Circuit and ErrorProbProtocol custom types.
    model_config = {"arbitrary_types_allowed": True, "frozen": True}

    # ====== Required Class Attributes =====

    # We assume a single instance of error model has a single error type and a single
    # application mode.
    # No initial value given, Pydantic will enforce the constructor to provide them.
    circuit: Circuit
    error_type: ErrorType  # Define the instruction type of the error model.
    is_time_dependent: bool
    application_mode: ApplicationMode

    # ====== Attribute ======

    # Dictionary of gate durations, mapping gate names to their execution times.
    # It must assign a duration to each gate type used in the circuit.
    # For time-independent models, this can be left undefined (None).
    gate_durations: Optional[dict[str, float]] = None

    def update_gate_durations(self, gate_durations: dict[str, float]):
        """Update the gate durations for the error model.
        This will recompute the operation times and tick durations if the model is
        time-dependent."""
        new_gate_durations = self.validate_gate_duration(gate_durations)
        return self.model_copy(
            update={
                "gate_durations": new_gate_durations,
            }
        )

    # Dictionary of gate error probabilities, mapping gate names to a callable
    # that returns the error probabilities for that gate, with optional time attribute.
    gate_error_probabilities: Optional[dict[str, GateErrorProbProtocol]] = defaultdict(
        lambda: lambda _: [0.0]
    )

    # If the application mode is END_OF_TICK or IDLE_END_OF_TICK, this callable defines
    # the error probability for a given time in the circuit.
    # It returns a list of floats according to the expected parameters of the
    # noise instruction.
    global_time_error_probability: ErrorProbProtocol = lambda _, __: [0.0]

    # If the model is time-dependent, this will be computed after initialization.
    # List the ticks duration, ordered according to occurance in the circuit.
    # !! Note: This contains the duration, not the time at which the tick occurs.
    # In order to compute the time at which the tick occurs,
    # you must sum the tick durations up to that point.
    _tick_durations: Optional[list[float]] = PrivateAttr(default=None)

    # If the model is time-dependent, this will be computed after initialization.
    # Dictionary mapping gate IDs to their execution times.
    _op_time: Optional[dict[str, float]] = PrivateAttr(default=None)

    # If the model depends on the idle time of channels during ticks,
    # this will be computed after initialization.
    # Dictionary mapping channel IDs to a list of idle times during each tick.
    _idle_times: Optional[dict[str, list[float]]] = PrivateAttr(default=None)

    def model_post_init(self, __context: Any) -> None:
        """Post-initialization method to compute operation times and ticks duration
        if the model is time-dependent.

        This is called after the model is initialized and all validators have run.
        If the model is time-dependent, it computes the operation times and tick
        durations. This bypass the frozen nature of the model to set the private
        attributes.
        """

        if self.is_time_dependent and self.gate_durations is not None:
            op_time, tick_durations, idle_durations = (
                self._compute__op_times_and__tick_durations()
            )
            object.__setattr__(self, "_idle_times", idle_durations)
            object.__setattr__(self, "_op_time", op_time)
            object.__setattr__(self, "_tick_durations", tick_durations)
        else:
            object.__setattr__(self, "_op_time", {})
            object.__setattr__(self, "_tick_durations", [])
            object.__setattr__(self, "_idle_times", {})

    # ====== Validators =====

    # if time-dependent, gate_durations must be provided
    @model_validator(mode="after")
    @classmethod
    def check_duration_defined_if_time_dependent(cls, model):
        """Ensure that gate_durations is defined if the model is time-dependent."""

        if not isinstance(model, CircuitErrorModel):
            raise TypeError("model must be an instance of CircuitErrorModel")
        if model.is_time_dependent and model.gate_durations is None:
            raise ValueError(
                "gate_durations must be provided for time-dependent error models."
            )
        return model

    @model_validator(mode="after")
    @classmethod
    def check_idle_application_is_time_dependent(cls, model):
        """Ensure that idle application mode is only used for time-dependent models."""
        if (
            model.application_mode == ApplicationMode.IDLE_END_OF_TICK
            and not model.is_time_dependent
        ):
            raise ValueError(
                "Idle application mode can only be used with time-dependent error "
                "models."
            )
        return model

    # if time-dependent, _op_time and tick_time must be computed
    @model_validator(mode="after")
    @classmethod
    # pylint: disable=protected-access
    def check__op_time_defined_if_time_dependent(cls, model):
        """Ensure that _op_time is defined if the model is time-dependent."""
        if model.is_time_dependent and model._op_time is None:
            raise ValueError(
                "_op_time must be computed for time-dependent error models."
            )
        if model.is_time_dependent and model._tick_durations is None:
            raise ValueError(
                "_tick_durations must be computed for time-dependent error models."
            )
        return model

    @model_validator(mode="after")
    @classmethod
    def validate_gate_duration(cls, model):
        """Ensure the gates duration are defined for all gate used in the circuit
        and that each gate has a valid duration.
        """
        unrolled = Circuit.unroll(model.circuit)
        gate_set = {op.name for layer in unrolled for op in layer if op is not None}
        v = model.gate_durations

        if v is not None:
            if not isinstance(v, dict):
                raise TypeError("gate_duration must be a dictionary or None")

            missing_gates = [gate for gate in gate_set if gate not in v]
            if missing_gates:
                raise ValueError(
                    f"Missing gate durations for: {missing_gates}."
                    f"Must provide durations, for all gates in {gate_set}",
                )
        return model

    @field_validator("gate_error_probabilities", mode="before")
    @classmethod
    def validate_gate_error_probabilities(
        cls, v: dict
    ) -> dict[str, GateErrorProbProtocol]:
        """
        Validate the gate error probabilities dictionary.
        Ensure it contains only gates in the gate_set
        Also ensure that each value is a callable.

        Parameters
        ----------
        v : dict[str, GateErrorProbProtocol] | None
            The dictionary of gate error probabilities, where keys are gate names
            and values are callables that return a list of floats representing the
            error probabilities for that gate.
            If None, it defaults to a callable that returns 0.0 for all gates.
        Returns
        -------
        dict[str, GateErrorProbProtocol]
            A validated dictionary of gate error probabilities, where each key is a
            gate name and each value is a callable that returns the
            error probabilities for that gate.
        Raises
        ------
        TypeError: If v is not a dictionary or callable.
        ValueError: If v contains invalid gate names or if a value is not callable.
        """

        # if undefined, return a default callable that returns 0.0
        # This can be allowed if end_of_tick application is used.
        if v is None:
            v = defaultdict(lambda: lambda _: [0.0])

        if not isinstance(v, dict):
            raise TypeError("gate_error_probabilities must be a dictionary")
        # Check it only contains valid gate names

        for key, func in v.items():
            # Check that it's callable
            if not callable(func):
                raise TypeError(f"Value for '{key}' must be callable.")
        invalid_values = [k for k, v in v.items() if not callable(v)]
        if invalid_values:
            raise TypeError(
                "All values in gate_error_probabilities must be callable. "
                f"Keys with invalid value: {invalid_values}"
            )

        # fill with 0 for missing gates
        return defaultdict(lambda: lambda _: [0.0], v)

    # Check that the global_time_error_probability is valid for t=0
    @field_validator("global_time_error_probability", mode="before")
    @classmethod
    def validate_global_time_error_probability(cls, v) -> ErrorProbProtocol:
        """Validate the tick error probabilities callable.
        Ensure it returns a float or a list of floats for t=0."""
        if v is None:
            # Default to single zero, will be reformatted by model validator if needed
            return lambda _, __: [0.0]
        if not callable(v):
            raise TypeError("global_time_error_probability must be a callable")
        try:

            def is_valid_prob(p):
                return isinstance(p, float) and 0.0 <= p <= 1.0

            result = v(0.0, 0.0)
            if isinstance(result, list):
                if not all(is_valid_prob(p) for p in result):
                    raise ValueError(
                        "global_time_error_probability must return a list of"
                        "floats between 0 and 1"
                    )
            else:
                raise TypeError(
                    "global_time_error_probability must return a list of floats"
                )
        except Exception as e:
            raise ValueError(
                f"global_time_error_probability callable failed: {e}"
            ) from e

        return v

    # ====== Validators that run last (to reformat default values) =====

    @model_validator(mode="after")
    @classmethod
    def validate_gate_error_probabilities_output(cls, model):
        """Validate that each lambda in gate_error_probabilities returns the correct
        number of parameters. The previous validators will typically assign default
        value of [0.0] to the gate error probabilities for each undefined gate,
        but some error types may require multiple parameters (e.g., PAULI_CHANNEL).
        This validator will reformat the default value to return a list of zeros with
        the correct length."""
        # Create a new dictionary to hold potentially modified functions
        updated_gate_error_probabilities = {}

        for gate_name, error_func in model.gate_error_probabilities.items():

            try:
                # Test the function with a dummy time value
                if model.is_time_dependent:
                    result = error_func(0.0)
                else:
                    result = error_func(None)

                # Check if this is a default value [0.0] that needs reformatting
                if (
                    len(result) == 1
                    and result[0] == 0.0
                    and model.error_type.param_count > 1
                ):
                    # Create a new function that returns the correct number of zeros
                    expected_count = model.error_type.param_count
                    # pylint: disable=cell-var-from-loop
                    if model.is_time_dependent:
                        updated_gate_error_probabilities[gate_name] = (
                            lambda t: [0.0] * expected_count
                        )
                    else:
                        updated_gate_error_probabilities[gate_name] = (
                            lambda _: [0.0] * expected_count
                        )
                else:
                    # Validate using the error type's validation method
                    model.error_type.validate_params(result)
                    updated_gate_error_probabilities[gate_name] = error_func

            except Exception as e:
                raise ValueError(
                    f"Gate '{gate_name}' error probability function failed validation: "
                    f"{e}"
                ) from e

        object.__setattr__(
            model,
            "gate_error_probabilities",
            defaultdict(
                lambda: lambda _: [0.0] * model.error_type.param_count,
                updated_gate_error_probabilities,
            ),
        )

        return model

    @model_validator(mode="after")
    @classmethod
    def validate_global_time_error_probability_output(cls, model):
        """Validate that global_time_error_probability returns the correct number of
        parameters. The previous validators will typically assign default value of [0.0]
        to the global time error probability, but some error types may require multiple
        parameters (e.g., PAULI_CHANNEL). This validator will reformat the default value
        to return a list of zeros with the correct length."""

        # Define a dummy zero value for tick_time and global_time in order to test the
        # function's behavior.
        tick_time = 0.0
        global_time = 0.0 if model.is_time_dependent else None
        try:
            result = model.global_time_error_probability(global_time, tick_time)

            # Check if this is a default value [0.0] that needs reformatting
            if (
                len(result) == 1
                and result[0] == 0.0
                and model.error_type.param_count > 1
            ):
                # Create a new function that returns the correct number of zeros
                expected_count = model.error_type.param_count
                # pylint: disable=unnecessary-lambda-assignment
                if model.is_time_dependent:
                    new_func = lambda t, t2: [0.0] * expected_count
                else:
                    new_func = lambda _, __: [0.0] * expected_count
                object.__setattr__(model, "global_time_error_probability", new_func)
            else:
                # Validate using the error type's validation method
                model.error_type.validate_params(result)

        except Exception as e:
            raise ValueError(
                f"global_time_error_probability function failed validation: {e}"
            ) from e

        return model

    # ====== Methods =====
    def get_idle_tick_error_probability(
        self, tick_index: int, channel_id: str
    ) -> list[float] | None:
        """
        Get the error probability based on the time a specific channel was idle during a
        tick.

        Parameters
        ----------
        tick_index : int
            The index of the tick for which to get the error probability.
        channel_id : str
            The ID of the channel for which to get the idle time error probability.

        Returns
        -------
        list[float] | None
            List of floats representing the error probabilities for the idle tick.
            If the application mode is not IDLE_END_OF_TICK, returns None.
        """
        if self.application_mode != ApplicationMode.IDLE_END_OF_TICK:
            return None

        if channel_id not in self._idle_times:
            raise ValueError(f"Channel {channel_id} not found in idle times mapping.")

        # time dependent, check the tick index
        if tick_index < 0 or tick_index >= len(self._tick_durations):
            raise IndexError("Tick index out of range.")
        # compute the time at which the tick occurs (time at the end of tick)

        idle_in_tick = self._idle_times[channel_id][tick_index]

        time = sum(self._tick_durations[: tick_index + 1])
        p = self.global_time_error_probability(time, idle_in_tick)
        # if the error probability is 0, we return None so that it gets ignored instead
        if not p or all(x == 0.0 for x in p):
            return None

        return p

    def get_tick_error_probability(self, tick_index: int = None) -> list[float] | None:
        """
        Get the error probability for a specific tick in the circuit.

        Parameters
        ----------
        tick_index : int
            The index of the tick for which to get the error probability.

        Returns
        -------
        list[float] | None
            List of floats representing the error probabilities for the tick.
            If the application mode is not END_OF_TICK, returns None.
        """
        if self.application_mode != ApplicationMode.END_OF_TICK:
            return None  # Only applicable for END_OF_TICK application mode

        if not self.is_time_dependent:
            p = self.global_time_error_probability(None, None)
        else:
            if tick_index is None:
                raise ValueError(
                    "tick_index must be provided for time-dependent models."
                )
            # time dependent, check the tick index
            if tick_index < 0 or tick_index >= len(self._tick_durations):
                raise IndexError("Tick index out of range.")
            # compute the time at which the tick occurs (time at the end of tick)
            time = sum(self._tick_durations[: tick_index + 1])
            p = self.global_time_error_probability(
                time, self._tick_durations[tick_index]
            )

        # if the error probability is 0, we return None so that it gets ignored instead
        if not p or all(x == 0.0 for x in p):
            return None

        return p

    def get_gate_error_probability(self, gate: Circuit) -> list[float] | None:
        """
        Get the error type (instruction) and probability for a given gate.

        Parameters
        ----------
        gate : Circuit
            The quantum gate for which to get the error type and probability.

        Returns
        -------
        list[float] | None
            List of floats representing the error probabilities parameters given
            the gate name
            if the error probability is 0, return None so that it gets ignored
            if the application mode is END_OF_TICK, return None
        """
        if not isinstance(gate, Circuit):
            raise TypeError("gate must be an instance of Circuit")
        if gate.circuit != ():
            raise ValueError("gate must have empty children (no sub-circuit)")

        if self.application_mode == ApplicationMode.END_OF_TICK:
            # If the application mode is END_OF_TICK, return None
            return None

        # pylint: disable=unsubscriptable-object
        if self.is_time_dependent:
            if gate.id not in self._op_time:
                raise ValueError(f"Gate {gate.id} not found in operation time mapping.")
            time = self._op_time.get(gate.id)
            if self.application_mode == ApplicationMode.AFTER_GATE:
                # If the error is applied after the gate, we use the time of the gate.
                time = self.gate_durations[gate.name] + time
            error_probability = self.gate_error_probabilities[gate.name](time)
        else:
            error_probability = self.gate_error_probabilities[gate.name](None)
        # if the error probability is 0, we return None so that it gets ignored instead
        # of  being converted to a noise instruction with 0 probability.
        if not error_probability or all(x == 0.0 for x in error_probability):
            return None

        return error_probability

    @property
    def total_time(self) -> float:
        """Total time of the circuit, computed as the sum of all tick durations.

        Returns
        -------
        float
            The total time of the circuit.

        Raises
        ------
        ValueError
            If the circuit is not time-dependent or tick durations are not set.
        """
        if not self.is_time_dependent or self._tick_durations is None:
            raise ValueError(
                "Circuit is not time-dependent or tick durations are not set."
            )

        return sum(self._tick_durations)

    # ====== Utilities ======
    # pylint: disable=unsubscriptable-object
    def _compute__op_times_and__tick_durations(
        self,
    ) -> tuple[dict[str, float], list[float], dict[str, list[float]]]:
        """Retrieve for each operation, the time elapsed from the start of the circuit
        until the start of the operation. Also computes the duration of each tick in the
        circuit and the idle time for each channel during each tick.

        Returns
        -------
        tuple[dict[str, float], list[float], dict[str, list[float]]]
            A tuple containing:

            - dict[str, float]: A dictionary mapping gate IDs to the execution times
              (time elapsed before the gate's execution starts).
            - list[float]: A list of tick durations for each layer in the circuit.
            - dict[str, list[float]]: A dictionary mapping channel IDs to a list of
              idle times during each tick.
        """
        _op_time = {}
        _tick_durations = []
        _channel_idle_duration_in_tick = defaultdict(lambda: [])
        time_stack = 0
        unrolled = Circuit.unroll(self.circuit)
        for tick in unrolled:
            channel_usage_time = defaultdict(lambda: 0)
            for operation in tick:
                # pylint: disable=unsupported-membership-test
                if operation.name not in self.gate_durations:
                    raise ValueError(
                        f"Gate {operation.name} not found in gate_durations dictionary."
                    )
                _op_time[operation.id] = time_stack
                for ch in operation.channels:
                    channel_usage_time[ch.id] += self.gate_durations[operation.name]
            _tick_duration = max(
                self.gate_durations[op.name] for op in tick if op is not None
            )
            _tick_durations.append(_tick_duration)
            # Store the idle time for each channel in the tick by subtracting the
            # total usage time from the tick duration.
            for ch in self.circuit.channels:
                if ch.is_quantum():
                    _channel_idle_duration_in_tick[ch.id].append(
                        _tick_duration - channel_usage_time[ch.id]
                    )
            time_stack += _tick_duration

        return _op_time, _tick_durations, _channel_idle_duration_in_tick


class HomogeneousTimeIndependentCEM(CircuitErrorModel):
    """
    A constant probability error model that applies a fixed error probability to all
    gates in the circuit.
    This model is not time-dependent, meaning the error probability does not change
    over time.

    Enforces the following properties:

    - The error model is not time-dependent.

    Parameters
    ----------
    circuit : Circuit
        The quantum circuit to which the error model will be applied.
    error_type : ErrorType
        The type of error that the model will apply to the circuit.
    application_mode : ApplicationMode
        The mode in which the error is applied to the circuit.
    error_probability : list[float]
        The error probability parameter(s) for the error model. This will be assigned
        to all target gates.
    target_gates : list[str]
        A list of gate names to which the error probability applies. Other gates will
        have an error probability of 0.0 by default.
    """

    is_time_dependent: bool = Field(
        init=False, default=False
    )  # This model is not time-dependent.

    # User need to define the error type and application mode
    error_type: ErrorType
    application_mode: ApplicationMode
    error_probability: float | list[float]
    global_time_error_probability: ErrorProbProtocol = Field(
        default=lambda _, __: [0.0], init=False
    )

    gate_error_probabilities: dict[str, GateErrorProbProtocol] | None = Field(
        default_factory=lambda: defaultdict(lambda: lambda _: [0]), init=False
    )

    @field_validator("error_probability")
    @classmethod
    def validate_error_probability(cls, v: float | list[float]) -> list[float]:
        """Validate the error probability to ensure it is a list of floats."""
        if isinstance(v, float):
            return [v]
        if isinstance(v, list) and all(isinstance(x, float) for x in v):
            return v
        raise ValueError("error_probability must be a float or a list of floats.")

    # Define the target gates to which the error probability applies. For the rest of
    # the gates, the error probability will be 0.0. By default, it applies to nothing
    target_gates: list[str] = []

    def model_post_init(self, __context):
        """
        Post-initialization for HomogeneousTimeIndependentCEM to set constant gate
        errors.
        """
        object.__setattr__(
            self,
            "gate_error_probabilities",
            self.validate_gate_error_probabilities(
                {g: lambda _: self.error_probability for g in self.target_gates}
            ),
        )
        # set the global time error probability to a constant value.
        object.__setattr__(
            self,
            "global_time_error_probability",
            lambda _, __: self.error_probability,  # error_probability is already a list
        )

        super().model_post_init(__context)


class HomogeneousTimeDependentCEM(CircuitErrorModel):
    """
    A constant probability error model that applies a fixed error probability function
    to all gates in the circuit.

    Enforces the following properties:

        - The error model is time-dependent.

    Parameters
    ----------
    error_type : ErrorType
        The type of error that the model will apply to the circuit.
    application_mode : ApplicationMode
        The mode in which the error is applied to the circuit.
    error_probability : ErrorProbProtocol
        The error probability parameter(s) for the error model. This will be assigned
        to all target gates.
    target_gates : list[str]
        A list of gate names to which the error probability applies. Other gates will
        have an error probability of 0.0 by default.
    """

    is_time_dependent: bool = Field(
        init=False, default=True
    )  # This model is time-dependent.
    # Define the error type and application mode
    error_type: ErrorType
    application_mode: ApplicationMode

    target_gates: list[str] = []

    error_probability: ErrorProbProtocol

    def model_post_init(self, __context):
        """
        Post-initialization for HomogeneousTimeDependentCEM to set constant gate errors.
        """

        object.__setattr__(
            self,
            "gate_error_probabilities",
            self.validate_gate_error_probabilities(
                {gate: self.error_probability for gate in self.target_gates}
            ),
        )
        # Set the global time error probability to a constant value according to the
        # given parameters.
        object.__setattr__(
            self,
            "global_time_error_probability",
            self.error_probability,
        )

        super().model_post_init(__context)


class AsymmetricDepolarizeCEM(CircuitErrorModel):
    """
    Error model that applies asymmetric depolarizing noise to the circuit.

    Enforce the following properties:
    - The error model is time-dependent.
    - The error type is PAULI_CHANNEL.
    - The application mode is END_OF_TICK.

    Parameters
    ----------
    t1 : float
        The time constant for the X and Y errors.
        Must be positive.
    t2 : float
        The time constant for the Z error.
        Must be positive.
    """

    t1: float
    t2: float
    gate_error_probabilities: dict[str, GateErrorProbProtocol] | None = Field(
        default_factory=lambda: defaultdict(lambda: lambda _: [0]), init=False
    )
    global_time_error_probability: ErrorProbProtocol = Field(
        default=lambda _, __: [0], init=False
    )

    @model_validator(mode="after")
    @classmethod
    def validate_time_constants(cls, model) -> float:
        """t2 must be smaller or equal 2*t1."""
        if model.t2 > 2 * model.t1:
            raise ValueError("t2 must be smaller or equal than 2*t1.")
        return model

    @field_validator("t1", "t2")
    @classmethod
    def check_positive(cls, v, info):
        """Ensure that t1 and t2 are positive."""
        if v < 0:
            raise ValueError(f"{info.field_name} must be non-negative.")
        return v

    def _p(self, t: float) -> list[float]:
        """
        Internal method to compute error probabilities given time t.
        Using the following model:
        p_x = p_y = (1 - exp(-t / t1)) / 4
        p_z = (1 - exp(-t / t2)) / 2 - p_x

        Parameters:
        ----------
        t : float
            The time used to compute the error probabilities.

        Returns:
        -------
            List[float]: [p_x, p_y, p_z]
        """
        if self.t1 == 0:
            exp_t_t1 = 0
        else:
            exp_t_t1 = math.exp(-t / self.t1)
        p_x = p_y = (1 - exp_t_t1) / 4
        if self.t2 == 0:
            exp_t_t2 = 0
        else:
            exp_t_t2 = math.exp(-t / self.t2)
        p_z = (1 - exp_t_t2) / 2 - p_x
        return [p_x, p_y, p_z]

    is_time_dependent: bool = Field(
        init=False, default=True
    )  # This model is time-dependent.
    error_type: ErrorType = Field(init=False, default=ErrorType.PAULI_CHANNEL)
    application_mode: ApplicationMode = Field(
        init=False, default=ApplicationMode.END_OF_TICK
    )

    def model_post_init(self, __context):
        """
        Post-initialization for AsymmetricDepolarizeCEM to set gate error probabilities.
        """
        object.__setattr__(
            self,
            "global_time_error_probability",
            lambda _, t: self._p(t),
        )
        super().model_post_init(__context)



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from loom.eka import Circuit, ChannelType, PauliOperator, Stabilizer
from loom.interpreter import InterpretationStep, Syndrome
from loom.interpreter.applicator import measureblocksyndromes, generate_syndromes
from loom.eka.utilities import (
    Direction,
    SingleQubitPauliEigenstate,
    ResourceState,
    Orientation,
)
from loom.eka.operations import MeasureBlockSyndromes, StateInjection

from ..code_factory import RotatedSurfaceCode


def state_injection(
    interpretation_step: InterpretationStep,
    operation: StateInjection,
    same_timeslice: bool,
    debug_mode: bool,
):
    """
    Inject a resource state into the specified block. This operation resets the central
    qubit of the block into the T state and resets the rest of the data qubits into four
    quadrants, such that two quadrants are in the $|0⟩$ state and two quadrants are in 
    the $|+⟩$ state. This ensures that the Z stabilizer measurements are deterministic 
    in the $|0⟩$ quadrants and the X stabilizer measurements are deterministic in the 
    $|+⟩$ quadrants.

    Note: the syndromes are measured once at the end of the operation in order to
    displace the logical operators to the top-left corner of the block.

    The algorithm is the following:

    - A.) Begin StateInjection composite operation session

    - B.) Circuit
        - B.1) Resets the central qubit in a physical T state.
        - B.2) Resets the rest of the data qubits into four quadrants, such that two \
        quadrants are in the $|0⟩$ state and two quadrants are in the $|+⟩$ state.
            
    - C.) Logical Operators
        - C.1) Displace the logical operators to the center of the block.
        
    - D.) Syndromes
        - D.1) Creates syndromes for the block after the T state injection. Only \
        the stabilizers that are deterministic generate syndromes.
            
    - E.) Measure Block Syndromes
        - E.1) Measure the block syndromes.
        
    - F.) Displace Logical Operators
        - F.1) Create the new logical operators on the top-left corner of the block.
        
    - G.) Final Circuit
        - G.1) End the composite operation session and append the full state injection \
        circuit

    Parameters
    ----------
    interpretation_step : InterpretationStep
        The interpretation step containing the input block.
    operation : StateInjection
        The operation containing the name of the input block to inject the T state into.
    same_timeslice : bool
        Whether to append the circuit to the same timeslice as the current
        interpretation step.
    debug_mode : bool
        Flag indicating whether the interpretation should be done in debug mode.
        Currently, the effects of debug mode are:
        - Disabling the commutation validation of Block
    """
    input_block = interpretation_step.get_block(operation.input_block_name)

    # A) Begin composite operation session
    interpretation_step.begin_composite_operation_session_MUT(
        same_timeslice=same_timeslice,
        circuit_name=(
            f"Inject {operation.resource_state.value} into block "
            f"{input_block.unique_label} and measure syndromes"
        ),
    )

    if not isinstance(input_block, RotatedSurfaceCode):
        raise TypeError(
            f"Expected input_block to be of type RotatedSurfaceCode, "
            f"but got {type(input_block)}."
        )
    if not all(s % 2 == 1 for s in input_block.size):
        raise ValueError(
            f"Expected input_block.size to be all odd, but got {input_block.size}."
        )

    # B) Circuit
    #   B.1) Resets the central qubit in a physical resource state.
    qubit_to_reset = (
        input_block.upper_left_qubit[0] + input_block.size[0] // 2,
        input_block.upper_left_qubit[1] + input_block.size[1] // 2,
        input_block.upper_left_qubit[2],
    )
    resource_state_circuit = get_physical_state_reset(
        interpretation_step, input_block, qubit_to_reset, operation.resource_state
    )

    #   B.2) Resets the rest of the data qubits into four quadrants
    reset_circuit, deterministic_stabs = reset_into_four_quadrants(
        interpretation_step,
        input_block,
    )

    state_injection_circuit = Circuit(
        name=(
            f"Inject {operation.resource_state.value} into block "
            f"{input_block.unique_label}"
        ),
        circuit=((resource_state_circuit, reset_circuit),),
    )
    interpretation_step.append_circuit_MUT(
        circuit=state_injection_circuit,
        same_timeslice=False,  # Prevent the same timeslice flag from being set
    )

    # C) Logical Operators
    # C.1) Displace the logical operators to the center of the block.
    centered_x_op, centered_z_op = find_centered_logical_operators(
        input_block=input_block, center_qubit=qubit_to_reset
    )

    centered_block = RotatedSurfaceCode(
        stabilizers=input_block.stabilizers,
        logical_x_operators=[centered_x_op],
        logical_z_operators=[centered_z_op],
        syndrome_circuits=input_block.syndrome_circuits,
        stabilizer_to_circuit=input_block.stabilizer_to_circuit,
        unique_label=input_block.unique_label,
        skip_validation=not debug_mode,
    )
    # update_evolution is set to False because the state injection resets the block into
    # a known state, so there is no evolution to record.
    interpretation_step.update_block_history_and_evolution_MUT(
        new_blocks=(centered_block,), old_blocks=(input_block,), update_evolution=False
    )

    # D) Syndromes
    # D.1) Create syndromes for the block after the state injection.
    deterministic_syndromes = create_deterministic_syndromes(
        interpretation_step=interpretation_step,
        block=centered_block,
        deterministic_stabs=deterministic_stabs,
    )
    interpretation_step.append_syndromes_MUT(syndromes=deterministic_syndromes)

    # E) Measure Block Syndromes
    # E.1) Measure the block syndromes.
    interpretation_step = measureblocksyndromes(
        interpretation_step=interpretation_step,
        operation=MeasureBlockSyndromes(
            input_block_name=centered_block.unique_label, n_cycles=1
        ),
        same_timeslice=False,  # Prevent the same timeslice flag from being set
        debug_mode=debug_mode,
    )

    # F) Displace Logical Operators
    # F.1) Create the new logical operators
    new_x_op, required_x_stabs = centered_block.get_shifted_equivalent_logical_operator(
        initial_operator=centered_block.logical_x_operators[0],
        new_upleft_qubit=centered_block.upper_left_qubit,
    )
    new_z_op, required_z_stabs = centered_block.get_shifted_equivalent_logical_operator(
        initial_operator=centered_block.logical_z_operators[0],
        new_upleft_qubit=centered_block.upper_left_qubit,
    )
    # Update the block with the new logical operators
    new_block = RotatedSurfaceCode(
        stabilizers=centered_block.stabilizers,
        logical_x_operators=[new_x_op],
        logical_z_operators=[new_z_op],
        syndrome_circuits=centered_block.syndrome_circuits,
        stabilizer_to_circuit=centered_block.stabilizer_to_circuit,
        unique_label=centered_block.unique_label,
        skip_validation=not debug_mode,
    )
    interpretation_step.update_block_history_and_evolution_MUT(
        new_blocks=(new_block,), old_blocks=(centered_block,)
    )
    # Add the required syndromes
    required_x_cbits = interpretation_step.retrieve_cbits_from_stabilizers(
        stabs_required=required_x_stabs, current_block=new_block
    )
    required_z_cbits = interpretation_step.retrieve_cbits_from_stabilizers(
        stabs_required=required_z_stabs, current_block=new_block
    )
    interpretation_step.update_logical_operator_updates_MUT(
        operator_type="X",
        logical_operator_id=new_x_op.uuid,
        new_updates=required_x_cbits,
        inherit_updates=True,
    )
    interpretation_step.update_logical_operator_updates_MUT(
        operator_type="Z",
        logical_operator_id=new_z_op.uuid,
        new_updates=required_z_cbits,
        inherit_updates=True,
    )

    # G) Final Circuit
    # G.1) End the composite operation session and append the full state injection
    # circuit
    state_injection_circuit = interpretation_step.end_composite_operation_session_MUT()
    interpretation_step.append_circuit_MUT(state_injection_circuit, same_timeslice)

    return interpretation_step


def get_physical_state_reset(
    interpretation_step: InterpretationStep,
    input_block: RotatedSurfaceCode,
    qubit_to_reset: tuple[int, int, int],
    resource_state: ResourceState,
):
    """
    Resets a physical qubit in the given resource state. The given qubit must be a data
    qubit in the input block.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        The interpretation step containing the input block.
    input_block : RotatedSurfaceCode
        The input block containing the data qubits.
    qubit_to_reset : tuple[int, int, int]
        The qubit to reset into the given state. It must be a data qubit in the input
        block.
    resource_state: ResourceState
        The resource state to reset the qubit into. Currently, only T and S states are
        supported.

    Returns
    -------
    Circuit
        A circuit that resets the specified qubit into the given resource state, which
        consists of a reset operation in $|+⟩$ followed by a physical gate that depends
        on the resource state.
    """
    if qubit_to_reset not in input_block.data_qubits:
        raise ValueError(
            f"Qubit {qubit_to_reset} is not a data qubit in the given block"
            f" {input_block.unique_label}."
        )

    qubit_channel = interpretation_step.get_channel_MUT(
        qubit_to_reset, channel_type=ChannelType.QUANTUM
    )
    match resource_state:
        case ResourceState.T:
            gate_sequence = (
                Circuit(name="Reset_+", channels=[qubit_channel]),
                Circuit(name="T", channels=[qubit_channel]),
            )
        case ResourceState.S:
            gate_sequence = (
                Circuit(name="Reset_+", channels=[qubit_channel]),
                Circuit(name="phase", channels=[qubit_channel]),
            )
        case _:
            raise ValueError(
                f"Resource state {resource_state} is not supported for physical reset."
            )

    return Circuit(
        name="Resource state reset",
        circuit=gate_sequence,
        channels=[qubit_channel],
    )


def find_qubits_quadrant(
    block: RotatedSurfaceCode,
    quadrant: Direction,
) -> tuple[tuple[int, int, int], ...]:
    """
    Finds the qubits in a specific quadrant of a rotated surface code block.

    Parameters
    ----------
    block : RotatedSurfaceCode
        The rotated surface code block containing the stabilizers.
    quadrant : Direction
        The direction of the quadrant to search for qubits.

    Returns
    -------
    tuple[tuple[int, int, int], ...]
        The qubits found in the specified quadrant.

    Raises
    ------
    ValueError
        If there is more than one pauli type in the boundary stabilizers.
    """
    # Order the stabilizers by the x coordinate for LEFT/RIGHT quadrants
    # and by the y coordinate for TOP/BOTTOM quadrants.
    # The reverse order is used for the RIGHT and BOTTOM quadrants.
    ordered_stabilizers = sorted(
        block.stabilizers,
        key=lambda stab: (
            stab.data_qubits[0][0]
            if quadrant in (Direction.LEFT, Direction.RIGHT)
            else stab.data_qubits[0][1]
        ),
        reverse=(quadrant in (Direction.RIGHT, Direction.BOTTOM)),
    )

    boundary_stabs = block.boundary_stabilizers(quadrant)
    pauli_quadrant = set(p for stab in boundary_stabs for p in stab.pauli)
    if len(pauli_quadrant) != 1:
        raise ValueError(
            f"Expected a single Pauli type for quadrant {quadrant},"
            f" found: {pauli_quadrant}"
        )

    # First find the qubits included in the quadrant because of the boundary stabilizers
    quadrant_qubit_set = set(
        q for stab in block.boundary_stabilizers(quadrant) for q in stab.data_qubits
    )
    # Iterate through the ordered stabilizers and populate the quadrant_qubit_set with
    # qubits that are part of the stabilizers with the same Pauli type
    # and two data qubits already added to the quadrant (e.g. the property of
    # stabilizers being deterministic propagates to the closest stabilizers from the
    # boundary).
    for stab in ordered_stabilizers:
        if (
            set(stab.pauli) == set(pauli_quadrant)
            and len(quadrant_qubit_set.intersection(stab.data_qubits)) == 2
        ):
            quadrant_qubit_set |= set(stab.data_qubits)

    return tuple(quadrant_qubit_set)


def reset_into_four_quadrants(
    interpretation_step: InterpretationStep,
    block: RotatedSurfaceCode,
) -> tuple[Circuit, tuple[Stabilizer, ...]]:
    """
    Reset the rest of the data qubits of the block into four quadrants, such that two
    quadrants are in the $|0⟩$ state, making the Z stabilizer measurements deterministic
    and two quadrants are in the $|+⟩$ state, making the X stabilizer measurements
    deterministic.

    E.g. for a distance 5 block, the reset will look like this:

    .. code-block::

                 d           d
        + --- 0 --- 0 --- 0 --- 0
      d |     |     |  d  |     |
        + --- + --- 0 --- 0 --- +
        |  d  |     |     |     | d
        + --- + --- 𝛙 --- + --- +
      d |     |     |     |  d  |
        + --- 0 --- 0 --- + --- +
        |     |  d  |     |     | d
        0 --- 0 --- 0 --- 0 --- +
           d           d

    "d" denotes a deterministic stabilizer measurement in the above figure.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        The interpretation step containing the input block.
    input_block : RotatedSurfaceCode
        The input block to reset.

    Returns
    -------
    tuple[Circuit, tuple[Stabilizer, ...]]
        The circuit resetting all qubits aside from the central qubit and the
        stabilizers that are initialized in a deterministic state because of these
        resets.
    """

    # Find the qubits in each quadrant and associate them with their reset state.
    qubits_to_state_map: dict[
        tuple[tuple[int, int, int], ...], SingleQubitPauliEigenstate
    ] = {}
    for quadrant in Direction:
        pauli_type = set(
            p for stab in block.boundary_stabilizers(quadrant) for p in stab.pauli
        )
        if len(pauli_type) != 1:
            raise ValueError(
                f"Expected a single Pauli type for quadrant {quadrant},"
                f" found: {pauli_type}"
            )
        quadrant_qubits = find_qubits_quadrant(block, quadrant)
        qubits_to_state_map |= {
            qubit: (
                SingleQubitPauliEigenstate.ZERO
                if pauli_type == {"Z"}
                else SingleQubitPauliEigenstate.PLUS
            )
            for qubit in quadrant_qubits
        }

    # Create the reset circuits:
    reset_circuit = Circuit(
        name="reset four quadrants",
        circuit=(
            (
                Circuit(
                    name=f"reset_{state.value}",
                    channels=[
                        interpretation_step.get_channel_MUT(
                            qubit, channel_type=ChannelType.QUANTUM
                        )
                    ],
                )
                for qubit, state in qubits_to_state_map.items()
            ),
        ),
    )

    deterministic_stabs = tuple(
        stab
        for stab in block.stabilizers
        if all(
            qubits_to_state_map.get(qubit, None) == ("+" if pauli == "X" else "0")
            for qubit, pauli in zip(stab.data_qubits, stab.pauli, strict=True)
        )
    )

    return reset_circuit, deterministic_stabs


def find_centered_logical_operators(
    input_block: RotatedSurfaceCode,
    center_qubit: tuple[int, int, int],
) -> tuple[PauliOperator, PauliOperator]:
    """
    Create new logical operators that go through the center of the block such that the
    state injection can be measured directly by the given logical operators.

    Parameters
    ----------
    input_block : RotatedSurfaceCode
        The input block that may or may not have centered logical operators.
    center_qubit : tuple[int, int, int]
        The qubit onto which the state is injected.

    Returns
    -------
    tuple[PauliOperator, PauliOperator]
        The centered logical operators.
    """
    x_log_is_horizontal = input_block.x_boundary == Orientation.HORIZONTAL
    x_op_len, z_op_len = (
        input_block.size if x_log_is_horizontal else input_block.size[::-1]
    )
    centered_logical_operators = (
        PauliOperator(
            pauli="X" * x_op_len,
            data_qubits=tuple(
                (
                    center_qubit[0] + i * x_log_is_horizontal,
                    center_qubit[1] + i * (not x_log_is_horizontal),
                    center_qubit[2],
                )
                for i in range(-(x_op_len // 2), x_op_len // 2 + 1)
            ),
        ),
        PauliOperator(
            pauli="Z" * z_op_len,
            data_qubits=tuple(
                (
                    center_qubit[0] + i * (not x_log_is_horizontal),
                    center_qubit[1] + i * (x_log_is_horizontal),
                    center_qubit[2],
                )
                for i in range(-(z_op_len // 2), z_op_len // 2 + 1)
            ),
        ),
    )

    return centered_logical_operators


def create_deterministic_syndromes(
    interpretation_step: InterpretationStep,
    block: RotatedSurfaceCode,
    deterministic_stabs: tuple[Stabilizer, ...],
) -> tuple[Syndrome, ...]:
    """
    Create syndromes that are deterministic for the given state.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        The interpretation step containing the input block.
    block : RotatedSurfaceCode
        The input block to create syndromes for.
    deterministic_stabs: tuple[Stabilizer, ...]
        The stabilizers that have a deterministic value after the state injection.

    Returns
    -------
    tuple[Syndrome, ...]
        The tuple of deterministic syndromes.
    """

    new_syndromes = generate_syndromes(
        interpretation_step=interpretation_step,
        stabilizers=deterministic_stabs,
        block=block,
        stab_measurements=tuple(() for _ in deterministic_stabs),
    )
    return new_syndromes



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from loom.eka import Circuit
from loom.interpreter import InterpretationStep

from .y_wall_out_utilities import (
    get_final_block_syndrome_measurement_cnots_circuit,
)
from ..move_block import (
    generate_syndrome_measurement_circuit_and_cbits,
    generate_and_append_block_syndromes_and_detectors,
)
from ...code_factory import RotatedSurfaceCode


def y_wall_out_final_qec_rounds(
    interpretation_step: InterpretationStep,
    block: RotatedSurfaceCode,
    same_timeslice: bool,
    debug_mode: bool,  # pylint: disable=unused-argument
) -> InterpretationStep:
    """
    Perform the final syndrome measurement rounds for the Y wall out operation.
    This operation performs d-2 rounds of syndrome measurement on the block in such
    a way that the final circuit is fault-tolerant.

    - A.) Begin composite operation for potentially multiple rounds

    - B.) Circuit generation
        - B.1) Generate ancilla initialization circuit
        - B.2) Generate CNOT circuit
        - B.3) Generate measurement circuit and classical bits
        - B.4) Assemble and append circuit

    - C.) Syndrome and detector generation
        - C.1) Generate and append block syndromes and detectors

    - D.) End the composite operation and append the circuit

    Parameters
    ----------
    interpretation_step : InterpretationStep
        The interpretation step.
    block : RotatedSurfaceCode
        The rotated surface code block to perform the final syndrome measurement rounds
        on.
    same_timeslice : bool
        Whether to append the final circuit in the same timeslice as the previous one.
    debug_mode : bool
        Whether to enable debug mode. (Unused since no new Blocks are created here)

    Returns
    -------
    InterpretationStep
        The updated interpretation step.
    """

    # Run d-2 rounds of syndrome measurement
    distance = min(block.size)
    n_total_rounds = distance - 2

    # A) Begin composite operation for potentially multiple rounds
    interpretation_step.begin_composite_operation_session_MUT(
        same_timeslice=same_timeslice,
        circuit_name=(
            f"final {n_total_rounds} syndrome measurement round(s) "
            f"on block {block.unique_label}"
        ),
    )

    for _ in range(n_total_rounds):
        # B) Circuit generation
        # B.1) Generate ancilla initialization circuit
        final_block_syndrome_measurement_reset_circuit = Circuit(
            name="Initialization of syndrome measurement ancilla",
            circuit=[
                [
                    Circuit(
                        f"reset_{'0' if stab.pauli_type == 'Z' else '+'}",
                        channels=[
                            interpretation_step.get_channel_MUT(stab.ancilla_qubits[0])
                        ],
                    )
                    for stab in block.stabilizers
                ]
            ],
        )

        # B.2) Generate CNOT circuit
        final_block_syndrome_measurement_cnot_circuit = (
            get_final_block_syndrome_measurement_cnots_circuit(
                block, interpretation_step
            )
        )

        # B.3) Generate measurement circuit and classical bits
        final_block_syndrome_measurement_measure_circuit, final_block_meas_cbits = (
            generate_syndrome_measurement_circuit_and_cbits(
                interpretation_step,
                block,
            )
        )

        # B.4) Assemble and append circuit
        interpretation_step.append_circuit_MUT(
            Circuit(
                name="one round of final syndrome measurement",
                circuit=Circuit.construct_padded_circuit_time_sequence(
                    (
                        (final_block_syndrome_measurement_reset_circuit,),
                        (final_block_syndrome_measurement_cnot_circuit,),
                        (final_block_syndrome_measurement_measure_circuit,),
                    )
                ),
            ),
            same_timeslice=False,
        )

        # C) Syndrome and detector generation
        generate_and_append_block_syndromes_and_detectors(
            interpretation_step, block, final_block_meas_cbits
        )

    # D) End the composite operation and append the circuit
    wrapped_circuit = interpretation_step.end_composite_operation_session_MUT()
    interpretation_step.append_circuit_MUT(wrapped_circuit, same_timeslice)

    return interpretation_step



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

# pylint: disable=too-many-locals, duplicate-code


from __future__ import annotations
from uuid import uuid4
from functools import cached_property
from pydantic.dataclasses import dataclass

from loom.eka import (
    Block,
    Lattice,
    LatticeType,
    PauliOperator,
    Stabilizer,
    SyndromeCircuit,
    Circuit,
    Channel,
)
from loom.eka.utilities import dataclass_config


@dataclass(config=dataclass_config)
class FiveQubitPerfectCode(Block):
    """
    Represents the five qubit perfect code, also known as the
    Laflamme code, as a subclass of ``Block``.
    """

    @classmethod
    def create(
        cls,
        lattice: Lattice,
        unique_label: str | None = None,
        position: tuple[int, int] = (0, 0),
        skip_validation: bool = False,
    ) -> FiveQubitPerfectCode:
        """Create a `Block` object for a five qubit perfect code/Laflamme code block.
        The five qubit code is a [[5, 1, 3]] quantum error-correcting code that can
        correct arbitrary single-qubit errors on any one qubit, and detect errors on any
        two qubits. There are 4 stabilizers of support 4 that are symmetric under cyclic
        permutation. The block is defined on a pentagonal lattice.

        NOTE: There are two stabilizers labelled with "XZZX" but with different data
        qubits. The two correspond to XZZXI and IXZZX stabilizers, but as the stabilizer
        class does not accept "I" as an input, these two become labelled identically.

        The circuit implementation is not fault tolerant, and is based on the stabilizer
        circuit from https://en.wikipedia.org/wiki/Five-qubit_error_correcting_code


        Parameters
        ----------
        lattice : Lattice
            Lattice on which the block is defined. The qubit indices depend on the type
            of lattice.
        unique_label : str, optional
            Label for the block. It must be unique among all blocks in the initial CRD.
            If no label is provided, a unique label is generated automatically using the
            uuid module.
        position : tuple[int, ...], optional
            Position of the top left corner of the block on the lattice,
            by default (0, 0).
        skip_validation : bool, optional
            Skip validation of the block object, by default False.

        Returns
        -------
        Block
            Block object for the five qubit perfect code
        """

        # Input validation
        if lattice.lattice_type != LatticeType.POLY_2D:
            raise ValueError(
                "The creation of five qubit perfect code blocks is "
                "currently only supported for 2D pentagonal lattices. Instead "
                f"the lattice is of type {lattice.lattice_type}."
            )
        if lattice.unit_cell_size != 9:
            raise ValueError(
                "The five qubit perfect code block requires a lattice with exactly 5 "
                f"qubits and 4 ancilla in each unit cell. Got {lattice.unit_cell_size} "
                "qubits and ancilla instead."
            )

        if not isinstance(position, tuple) or any(
            not isinstance(x, int) for x in position
        ):
            raise ValueError(
                f"`position` must be a tuple of integers. Got '{position}' instead."
            )

        if unique_label is None:
            unique_label = str(uuid4())

        # Define the stabilizers
        looped_stabilizer_supports = [(0, 0, i % 5) for i in range(8)]
        stab_labels = ["XZZX", "XZZX", "XXZZ", "ZXXZ"]
        stabilizers = [
            Stabilizer(
                pauli=stab_labels[i],
                data_qubits=sorted(looped_stabilizer_supports[i : i + 4]),
                ancilla_qubits=[(0, 0, 5 + i)],
            )
            for i in range(4)
        ]

        # Define the logical operators
        all_data_qubits = [(0, 0, i) for i in range(5)]

        logical_x_operator = PauliOperator(
            pauli="X" * 5,
            data_qubits=all_data_qubits,
        )
        logical_z_operator = PauliOperator(
            pauli="Z" * 5,
            data_qubits=all_data_qubits,
        )

        # Define the syndrome extraction circuits
        syndrome_circuits = [
            cls.generate_syndrome_extraction_circuits("XZZXI"),
            cls.generate_syndrome_extraction_circuits("IXZZX"),
            cls.generate_syndrome_extraction_circuits("XIXZZ"),
            cls.generate_syndrome_extraction_circuits("ZXIXZ"),
        ]

        # Define the stabilizer to syndrome mapping
        stabilizer_to_circuit = {
            stab.uuid: syndrome_circuits[i].uuid for i, stab in enumerate(stabilizers)
        }

        block = cls(
            unique_label=unique_label,
            stabilizers=stabilizers,
            logical_x_operators=[logical_x_operator],
            logical_z_operators=[logical_z_operator],
            syndrome_circuits=syndrome_circuits,
            stabilizer_to_circuit=stabilizer_to_circuit,
            skip_validation=skip_validation,
        )
        if position == (0, 0):
            return block

        return block.shift(position)

    @staticmethod
    def generate_syndrome_extraction_circuits(pauli: str) -> SyndromeCircuit:
        """Generate syndrome extraction circuits for the five qubit perfect code.

        Parameters
        ----------
        pauli : str
            The Pauli operator for which to generate the syndrome extraction circuits.

        Returns
        -------
        SyndromeCircuit
            A SyndromeCircuit object containing the circuit for syndrome extraction.
        """

        # Stabilizer info
        p_index = pauli.index("I")
        start_t_index = (p_index + 1) % 5
        name = f"{pauli}_syndrome_extraction"

        # Define Channels
        data_channels = [Channel(type="quantum", label=f"d{i}") for i in range(4)]
        anc_channel = [Channel(type="quantum", label="a0")]
        cbit_channel = [Channel(type="classical", label="c0")]

        # Define the circuit
        reset_layer = [Circuit(name="Reset_0", channels=anc_channel)]

        hadamard_initial = [Circuit(name="H", channels=anc_channel)]
        hadamard_final = [Circuit(name="H", channels=anc_channel)]

        measurement_layer = [
            Circuit(name="Measurement", channels=anc_channel + cbit_channel)
        ]

        insert_entangling_layer = []
        i = 0  # len(pauli) == 5, so we track index for data channels
        for p in pauli:
            if p == "I":
                insert_entangling_layer.append([])
            else:
                insert_entangling_layer.append(
                    [Circuit(name=f"C{p}", channels=anc_channel + [data_channels[i]])]
                )
                i += 1

        entangling_layer = [[], [], []]
        entangling_layer = (
            entangling_layer[:start_t_index]
            + insert_entangling_layer
            + entangling_layer[start_t_index:]
        )

        circuit_list = (
            [reset_layer, hadamard_initial]
            + entangling_layer
            + [hadamard_final, measurement_layer]
        )

        # Return the syndrome extraction circuit
        return SyndromeCircuit(
            pauli=pauli[:p_index] + pauli[p_index + 1 :],
            name=name,
            circuit=Circuit(
                name=name,
                circuit=circuit_list,
                channels=data_channels + anc_channel + cbit_channel,
            ),
        )

    # Instance methods
    def __eq__(self, other: FiveQubitPerfectCode) -> bool:
        if not isinstance(other, FiveQubitPerfectCode):
            raise NotImplementedError(
                f"Cannot compare FiveQubitPerfectCode with {type(other)}"
            )
        return super().__eq__(other)

    def shift(
        self, position: tuple[int, ...], new_label: str | None = None
    ) -> FiveQubitPerfectCode:
        return super().shift(position, new_label)

    def rename(self, name: str) -> FiveQubitPerfectCode:
        return super().rename(name)

    @cached_property
    def stabilizers_labels(self) -> dict[str, dict[str, tuple[int, ...]]]:
        return super().stabilizers_labels



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from enum import Enum
import itertools

from pydantic.dataclasses import dataclass
from pydantic import field_validator, ValidationInfo
import numpy as np

from .utilities import (
    retrieve_field,
    dataclass_config,
)


class LatticeType(str, Enum):
    """Defines the types of lattices available."""

    LINEAR = "linear"
    TRIANGLE_2D = "triangle_2d"
    SQUARE_2D = "square_2d"
    SQUARE_2D_0ANC = "square_2d_0anc"
    SQUARE_2D_2ANC = "square_2d_2anc"
    CAIRO_PENT_2D = "cairo_pent_2d"
    HEX_2D = "hex_2d"
    OCT_2D = "oct_2d"
    POLY_2D = "poly_2D"
    CUBE_3D = "cube_3d"
    CUSTOM = "custom"


@dataclass(config=dataclass_config)
class Lattice:
    """
    A lattice object contains information about the lattice structure. The lattice
    defines the indexing system of qubits, since qubits are indexed by their coordinates
    in terms of the lattice and basis vectors.

    Note that this lattice on which stabilizers, logical qubit blocks, and lattice
    surgery operations are defined, is a priori only an abstract lattice which makes
    defining all these things and handling qubit indices convenient. It does not have to
    be the actual lattice of the physical hardware.

    Qubit indices have the form `(x_0, ..., x_{n-1}, b)` where `x_0, ..., x_{n-1}`
    denote the unit cell with `n` the dimension of the lattice and `b` denotes which
    qubit in the unit cell it is.

    E.g. for a 2D lattice with a 2-qubit unit cell, the qubit indices are `(x, y, b)`
    which means that it is the `x`-th unit cell along the first lattice vector, the
    `y`-th unit cell along the second lattice vector, and the `b`-th qubit in the unit
    cell (b is either 0 or 1 in the case of a 2-qubit unit cell).

    Parameters
    ----------
    basis_vectors : tuple[tuple[float, ...], ...]
        Tuple of basis vectors that define the unit cell. If there is more than one
        qubit in the unit cell, one has to specify their position inside the unit cell
        using the basis vectors.
    lattice_vectors : tuple[tuple[float, ...], ...]
        Tuple of lattice vectors that define the lattice. The whole lattice is obtained
        by translating the unit cell along these lattice vectors by integer amounts.
    size : tuple[int, ...] | None
        The size of the lattice in each dimension. If set to None, the lattice is
        assumed to be infinitely large.
    lattice_type : LatticeType
        Type of the lattice. This is useful for storing the lattice as well as for some
        function such as `Block` creation which behave differently for different lattice
        types. Default is LatticeType.CUSTOM which means the lattice is created by the
        user.
    """

    basis_vectors: tuple[tuple[float, ...], ...]
    lattice_vectors: tuple[tuple[float, ...], ...]
    size: tuple[int, ...] | None = None
    lattice_type: LatticeType = LatticeType.CUSTOM

    @property
    def n_dimensions(self) -> int:
        """
        The dimension of the lattice, i.e. the number of lattice vectors.
        """
        return len(self.lattice_vectors)

    @property
    def unit_cell_size(self) -> int:
        """
        The size of the unit cell, i.e. the number of basis vectors.
        """
        return len(self.basis_vectors)

    # Validation
    @field_validator("basis_vectors", mode="before")
    @classmethod
    def format_basis_vectors(cls, basis_vectors: tuple):
        """
        Format the basis_vector input to also accept [] for 1-qubit unit cells.
        """
        if len(basis_vectors) == 0:
            return [[0, 0]]
        return basis_vectors

    @field_validator("basis_vectors", mode="after")
    @classmethod
    def basis_vectors_same_length(cls, basis_vectors: tuple):
        """
        Validate that all basis vectors have the same length.
        """
        if any(
            len(basis_vector) != len(basis_vectors[0]) for basis_vector in basis_vectors
        ):
            raise ValueError("All basis vectors must have the same length.")
        return basis_vectors

    @field_validator("lattice_vectors", mode="after")
    @classmethod
    def lattice_vectors_same_length(cls, lattice_vectors: tuple):
        """
        Validate that all lattice vectors have the same length.
        """
        if any(
            len(lattice_vector) != len(lattice_vectors[0])
            for lattice_vector in lattice_vectors
        ):
            raise ValueError("All lattice vectors must have the same length.")
        return lattice_vectors

    @field_validator("size", mode="after")
    @classmethod
    def size_right_dimension(cls, size: int | None, values: ValidationInfo):
        """
        Validate that `size` has the right dimension, i.e. the size of the tuple has
        to be equal to the number of lattice vectors.
        """
        # Allow infinite lattices where size is set to None
        if size is None:
            return size

        # If the lattice is not infinite, check that the size has the right dimension
        n_dim = len(retrieve_field("lattice_vectors", values))
        if len(size) != n_dim:
            raise ValueError(
                f"The given `size` is invalid. `size` has {len(size)} "
                f"elements, but the lattice has {n_dim} dimensions."
            )
        return size

    @field_validator("size", mode="after")
    @classmethod
    def size_not_negative(cls, size):
        """Validate that `size` is not negative."""
        if size is not None and any(x < 0 for x in size):
            raise ValueError("Size cannot be negative.")
        return size

    # Methods
    def all_unit_cells(
        self, size: tuple[int | None, ...] | None = None
    ) -> list[tuple[int, ...]]:
        """
        Get a list of all unit cells of the lattice. They are given in terms of the
        lattice vectors.

        Parameters
        ----------
        size : tuple[int  |  None, ...] | None, optional
            If only a part of the lattice is needed, the size of the wanted section can
            be specified. If None is provided, all unit cells of the lattice will be
            returned. If the lattice has infinite size, the `size` parameter must be
            provided.

        Returns
        -------
        list[tuple[int, ...]]
            Unit cells of the lattice for the given dimensions. They are given as a list
            of tuples where each tuple contains the coordinates of the unit cell in
            terms of the lattice vectors.
        """
        # Validation
        if size is None and self.size is None:
            raise ValueError("Please specify the `size` parameter.")
        if self.size is None and any(x is None for x in size):
            raise ValueError("Please specify all dimensions in the `size` parameter.")
        if size is not None and len(size) != self.n_dimensions:
            raise ValueError(
                f"The given `size` is invalid. `size` has {len(size)} "
                f"elements, but the lattice has {self.n_dimensions} dimensions."
            )

        # Combine the size of the lattice with the given size
        if size is None:
            adapted_size = self.size
        else:
            adapted_size = [
                size[i] if size[i] is not None else self.size[i]
                for i in range(self.n_dimensions)
            ]
        # Get all unit cells inside the specified region
        grids = np.meshgrid(*[range(x) for x in adapted_size], indexing="ij")
        return list(zip(*[map(int, grid.flatten()) for grid in grids], strict=True))

    def all_qubits(
        self,
        size: tuple[int | None, ...] | None = None,
        force_including_basis: bool = False,
    ) -> list[tuple[int, ...]]:
        """
        Get a list of qubits in the lattice inside the region specified by `size`.

        Parameters
        ----------
        size : tuple[int  |  None, ...] | None, optional
            If only a part of the lattice is needed, the size of the wanted section can
            be specified. If None is provided, all qubits of the lattice will be
            returned. If the lattice has infinite size, the `size` parameter must be
            provided.
        force_including_basis : bool, optional
            By default, the basis parameter is included in the qubit indices only if
            `size > 1`. To force it to appear for `size == 1`, set this boolean to
            `True`.

        Returns
        -------
        list[tuple[int, ...]]
            List of all qubits inside the specified region.
        """
        unit_cells = self.all_unit_cells(size)
        if self.unit_cell_size == 1 and force_including_basis is False:
            return unit_cells

        bases_in_unit_cell = range(self.unit_cell_size)
        all_combinations = [
            tuple_ + (int_,)
            for tuple_, int_ in itertools.product(unit_cells, bases_in_unit_cell)
        ]
        return all_combinations

    # Default lattices

    @staticmethod
    def _points_on_circle(
        n_points: int,
        radius: float | int,
        offset: float | int = -0.5 * np.pi,
        disp: tuple[int | float, int | float] = (0, 0),
    ) -> list[list[float]]:
        """Generate points on a circle of given radius and number of points. Default
        offset places first point at the top of the circle for polygons with odd number
        of points, and disp shifts the points.

        Parameters
        ----------
        radius : float | int
            Radius of the circle on which the points are placed.
        n_points : int
            Number of points to be placed on the circle.
        offset : float | int, optional
            Offset to be applied to the angle of the points on the circle. This is
            useful for placing the first point at a specific angle. The default values
            for this parameter are accurate to the plotting convention of
            StabilizerPlot(), which inverts the y-axis:
                top of circle = -0.5 * np.pi
                right of circle = 0
                bot of circle = +0.5 * np.pi
                etc.
        disp : tuple[int | float, int | float], optional
            Displacement of the points from the center of the circle. This is useful
            for placing the points at a specific position in the 2D plane.

        Returns
        -------
        list[list[float]]
            List of points on the circle, each point represented as a list of two
            floats [x, y] coordinates.
        """

        # Validation
        if not isinstance(n_points, int) or n_points < 1:
            raise ValueError(
                "Number of points must be an integer that is at least 1. "
                f"Received {n_points} of type {type(n_points)}."
            )
        if not isinstance(radius, (int, float)) or radius <= 0:
            raise ValueError(
                "Radius must be a positive number (int or float). "
                f"Received {radius} of type {type(radius)}."
            )
        if not isinstance(offset, (int, float)):
            raise TypeError(
                "Offset must be a number (int or float). "
                f"Received {offset} of type {type(offset)}."
            )
        if (
            not isinstance(disp, tuple)
            or not all(isinstance(x, (int, float)) for x in disp)
            or len(disp) != 2
        ):
            raise TypeError(
                "Disp must be a tuple of two numbers (x, y). "
                f"Received {disp} of type {type(disp)}."
            )

        if n_points == 1:
            # If only one point is requested, return it at the specified displacement
            return [[disp[0], disp[1]]]

        return [
            [
                float(radius * np.cos(2 * np.pi * i / n_points + offset) + disp[0]),
                float(radius * np.sin(2 * np.pi * i / n_points + offset) + disp[1]),
            ]
            for i in range(n_points)
        ]

    @classmethod
    def linear(
        cls,
        lattice_size: tuple[int] | None = None,
    ):
        """Generates a linear lattice in 1D. The lattice is defined by a single lattice
        vector that is parallel to the x-axis. The unit cell is defined by set of two
        basis vectors. The unit cell contains 2 points (1 qubit, 1 ancilla) arranged on
        a straight line. The lattice vector tiles the plane in a linear pattern.

        Parameters
        ----------
        lattice_size : tuple[int] | None, optional
            The size of the lattice in terms of unit cells. If None, the lattice is
            assumed to be infinitely large.

        Returns
        -------
        Lattice
            A Lattice object representing the linear lattice.
        """
        lattice_type = LatticeType.LINEAR
        # Vectors are artificially expanded to 2D for plotting purposes
        basis_vectors = [[0, 0], [0.5, 0]]
        lattice_vectors = [[1, 0]]
        return cls(basis_vectors, lattice_vectors, lattice_size, lattice_type)

    @classmethod
    def triangle_2d(
        cls,
        lattice_size: tuple[int, int] | None = None,
    ):
        """Generates a triangular lattice in 2D. The lattice tiles the plane, and is
        defined by two lattice vectors that are 60 degrees to each other. The unit cell
        is defined by a set of three basis vectors. The unit cell contains 3 points
        (1 qubits, 2 ancilla) arranged in a downwards sloping line. The lattice vectors
        tile the plane in a triangular pattern.

        Basis vector 0 represents the data qubit. Basis vector 1 represents ancilla
        qubit for triangles with vertices pointing downwards, and basis vector 2
        represents ancilla qubit for triangles with vertices pointing upwards.

        Parameters
        ----------
        lattice_size : tuple[int, int] | None, optional
            The size of the lattice in terms of unit cells. If None, the lattice is
            assumed to be infinitely large.

        Returns
        -------
        Lattice
            A Lattice object representing the triangular lattice.
        """
        lattice_type = LatticeType.TRIANGLE_2D
        basis_vectors = [[0, 0], [0.5, 1 / (2 * np.sqrt(3))], [1, 1 / np.sqrt(3)]]
        lattice_vectors = [[1, 0], [0.5, 0.5 * np.sqrt(3)]]
        return cls(basis_vectors, lattice_vectors, lattice_size, lattice_type)

    @classmethod
    def square_2d_0anc(cls, lattice_size: tuple[int, int] | None = None):
        """Generates a square lattice in 2D. The lattice tiles the plane, and is
        defined by two lattice vectors that are orthogonal to each other. The unit
        cell is defined by one basis vector. The unit cell contains 1 point (1 qubit).
        The lattice vectors tile the plane in a square pattern.

        Basis vector 0 represents the data qubit.

        NOTE: This lattice was defined for the specific use of the HGP code class,
        because the product structure automatically allocates two dimensional
        coordinates to all qubits in the code block (both datas and checks)

        Parameters
        ----------
        lattice_size : tuple[int, int] | None, optional
            The size of the lattice in terms of unit cells. If None, the lattice is
            assumed to be infinitely large.

        Returns
        -------
        Lattice
            A Lattice object representing the square lattice.
        """
        lattice_type = LatticeType.SQUARE_2D_0ANC
        basis_vectors = [[0, 0]]
        lattice_vectors = [[1, 0], [0, 1]]
        return cls(basis_vectors, lattice_vectors, lattice_size, lattice_type)

    @classmethod
    def square_2d(
        cls,
        lattice_size: tuple[int, int] | None = None,
    ):
        """Generates a square lattice in 2D. The lattice tiles the plane, and is
        defined by two lattice vectors that are orthogonal to each other. The unit
        cell is defined by a set of two basis vectors. The unit cell contains 2 points
        (1 qubits, 1 ancilla) arranged in a downwards sloping line. The lattice vectors
        tile the plane in a square pattern.

        Basis vector 0 represents the data qubit, and basis vector 1 represents the
        ancilla qubit.

        Parameters
        ----------
        lattice_size : tuple[int, int] | None, optional
            The size of the lattice in terms of unit cells. If None, the lattice is
            assumed to be infinitely large.

        Returns
        -------
        Lattice
            A Lattice object representing the square lattice.
        """
        lattice_type = LatticeType.SQUARE_2D
        basis_vectors = [[0, 0], [-0.5, -0.5]]
        lattice_vectors = [[1, 0], [0, 1]]
        return cls(basis_vectors, lattice_vectors, lattice_size, lattice_type)

    @classmethod
    def square_2d_2anc(cls, lattice_size: tuple[int, int] | None = None, shift=0.15):
        """Generates a square lattice in 2D. The lattice tiles the plane, and is
        defined by two lattice vectors that are orthogonal to each other. The unit
        cell is defined by a set of two basis vectors. The unit cell contains 3 points
        (1 qubits, 2 ancilla) arranged such that the ancillas lie on the same
        line, but with a slight separation. The lattice vectors tile the plane in a
        square pattern.
        Basis vector 0 represents the data qubit, and basis vectors 1 and 2 represent
        the first ancilla and second ancilla qubit.

        NOTE: This lattice was defined for the specific use of the 488 Color code, where
        the ancillas are arranged to be slightly shifted from the center of a polygon
        and where the datas are placed at the vertices of the given polygon.

        Parameters
        ----------
        lattice_size : tuple[int, int] | None, optional
            The size of the lattice in terms of unit cells. If None, the lattice is
            assumed to be infinitely large.
        shift : float, optional
            The shift of the ancilla qubits from the center of the unit cell.
            Default is 0.15.
        Returns
        -------
        Lattice
            A Lattice object representing the square lattice.
        """
        lattice_type = LatticeType.SQUARE_2D_2ANC
        basis_vectors = [[0, 0], [0, -shift], [0, shift]]
        lattice_vectors = [[1, 0], [0, 1]]
        return cls(basis_vectors, lattice_vectors, lattice_size, lattice_type)

    @classmethod
    def cairo_pent_2d(
        cls,
        lattice_size: tuple[int, int] | None = None,
    ):
        """Generates a catalan pentagonal lattice which is a regular tiling of the
        plane. Although there are an infinite number of ways to tile the plane with
        pentagons, this implementation uses a bilaterally symmetric pentagon with 4
        long edges and 1 short edge in the ratio of 1 : sqrt(3) - 1. The angles of
        this pentagon (starting from top and moving clockwise) are 120, 90, 120, 120,
        and 90 degrees.

        Each unit cell contains 6 points. The first 5 points are the vertices of a
        pentagon, and the 6th point is a point on another pentagon that is adjacent to
        the first one. The points are arranged such that they tile the plane via the
        lattice vectors, though the tiling will not be intuitive due to the current
        unit cell definition.
        Returns
        -------
        Lattice
            A Lattice object representing the catalan pentagonal lattice.
        """

        lattice_type = LatticeType.CAIRO_PENT_2D
        basis_vectors = [
            [0, 0],
            [0.5 * np.sqrt(3), 0.5],
            [0.5 * (np.sqrt(3) - 1), 0.5 * (1 + np.sqrt(3))],
            [0.5 * (1 - np.sqrt(3)), 0.5 * (1 + np.sqrt(3))],
            [-0.5 * np.sqrt(3), 0.5],
            [-np.sqrt(3), 1],
        ]
        lattice_vectors = [[2 * np.sqrt(3), 0], [np.sqrt(3), np.sqrt(3)]]
        return cls(basis_vectors, lattice_vectors, lattice_size, lattice_type)

    @classmethod
    def hex_2d(
        cls,
        lattice_size: tuple[int, int] | None = None,
    ):
        """Generates a hexagonal lattice in 2D. The lattice tiles the plane, and is
        defined by two lattice vectors that are 60 degrees to each other. The unit
        cell contains 2 points (2 qubits) arranged in a downwards sloping line.
        The lattice vectors tile the plane in a hexagonal pattern.

        Parameters
        ----------
        lattice_size : tuple[int, int] | None, optional
            The size of the lattice in terms of unit cells. If None, the lattice is
            assumed to be infinitely large.

        Returns
        -------
        Lattice
            A Lattice object representing the hexagonal lattice.
        """

        lattice_type = LatticeType.HEX_2D
        basis_vectors = [[0, 0], [0, 1 / np.sqrt(3)]]
        lattice_vectors = [[1, 0], [0.5, 0.5 * np.sqrt(3)]]
        return cls(basis_vectors, lattice_vectors, lattice_size, lattice_type)

    @classmethod
    def oct_2d(
        cls,
        lattice_size: tuple[int, int] | None = None,
        r: float = 0.5,
        anc: int = 0,
    ):
        """Generates a semi-regular octagonal lattice with a specified radius `r` for
        the points on the circle. The lattice tiles the plane. Spaces between adjacent
        octagons are tiled by squares.

        Each unit cell contains 8 qubits + 4 sets of ancilla. The number of ancilla in
        each set is determined by the `anc` parameter, and is 0 by default. The data
        qubits are placed on a circle of radius `r`. The ancilla qubits are placed on a
        circle that is 1 / 2 the radius of the squares.

        First set with indices [8 : 8 + anc]:
            Ancilla for primary octagon. Placed on a circle of radius
            `r * sin(pi / 8) / 2`.
        Second set with indices [8 + anc : 8 + 2 * anc]:
            Ancilla for square below primary octagon. Placed on a circle of the same
            radius, shifted by half the distance between adjacent unit cells in the
            y-direction.
        Third set with indices [8 + 2 * anc : 8 + 3 * anc]:
            Ancilla for square to right of primary octagon. Placed on a circle of the
            same radius, shifted by half the distance between adjacent unit cells in
            the y-direction.
        Fourth set with indices [8 + 3 * anc : 8 + 4 * anc]:
            Ancilla for secondary octagon, diagonal down from primary octagon. Placed
            on a circle of the same radius, shifted by half the distance between
            adjacent unit cells in both the x and y directions.

        Parameters
        ----------
        lattice_size : tuple[int, int] | None, optional
            The size of the lattice in terms of unit cells. If None, the lattice is
            assumed to be infinitely large.
        r : float, optional
            The radius of the circle on which the points are placed. Default is 0.5.
        anc : int, optional
            The number of ancilla qubits in the unit cell. Default is 0. If anc is set
            to 1, ancilla will be placed at the centre of each octagonal and square
            tiles.

        Returns
        -------
        Lattice
            A Lattice object representing the octagonal lattice.
        """
        # Validation
        if not isinstance(anc, int) or anc < 0:
            raise ValueError("Number of ancilla qubits must be a non-negative integer.")

        lattice_type = LatticeType.OCT_2D
        anc_radius = r * np.sin(np.pi / 8) / 2
        cell_dist = (
            2 * r * (np.sin(np.pi / 8) + np.cos(np.pi / 8))
        )  # Ensures squares fit between octagons
        basis_vectors = cls._points_on_circle(8, r, -5 * np.pi / 8)

        if anc > 0:
            basis_vectors = basis_vectors + (
                cls._points_on_circle(anc, anc_radius)
                + cls._points_on_circle(anc, anc_radius, disp=(0, 0.5 * cell_dist))
                + cls._points_on_circle(anc, anc_radius, disp=(0.5 * cell_dist, 0))
                + cls._points_on_circle(
                    anc, anc_radius, disp=(0.5 * cell_dist, 0.5 * cell_dist)
                )
            )
        lattice_vectors = [[0, cell_dist], [cell_dist, 0]]
        return cls(basis_vectors, lattice_vectors, lattice_size, lattice_type)

    @classmethod
    def poly_2d(  # pylint:disable=too-many-arguments, too-many-positional-arguments
        cls,
        lattice_size: tuple[int, int] | None = None,
        n: int | None = None,
        poly_radius: float | int = 0.5,
        poly_offset: float | int = -0.5 * np.pi,
        cell_dist_factor: float | int = (1 + np.sqrt(3)),
        anc: int = 1,
        anc_radius: float | int = 0.2,
        anc_offset: float | int = -0.5 * np.pi,
        anc_disp: tuple[int | float, int | float] = (0, 0),
    ):
        """
        Generates a lattice of n-sided polygons. The points are placed on a circle
        of radius `poly_radius` and arranged in a polygonal pattern. The lattice does
        not tile the plane, but is useful for certain applications where polygonal
        symmetry is desired.

        Lattice vectors are defined such that the distance between adjacent unit cells
        is `poly_radius * cell_dist_factor`. Each unit cell contains n + anc points
        (`n` qubits, `anc` ancilla). The `anc` points are placed on a circle of radius
        `anc_radius`. If `anc` is set to 1, ancilla will be placed at the center of the
        polygon.

        Parameters
        ----------
        lattice_size : tuple[int, int] | None, optional
            The size of the lattice in terms of unit cells. If None, the lattice is
            assumed to be infinitely large.
        n : int, optional
            The number of sides of the polygon. Default is 5 (pentagon).
        poly_radius : float | int, optional
            The radius of the circle on which the points are placed. Default is 0.5.
        poly_offset : float | int, optional
            The offset to be applied to the angle of the points on the circle. This
            is useful for placing the first point at a specific angle. Default is
            -0.5 * np.pi (top of circle).
        cell_dist_factor : float | int, optional
            The factor by which the distance between adjacent unit cells is multiplied.
            This is useful for adjusting the distance between the polygons. Default is
            (1 + np.sqrt(3)) as a purely aesthetic choice to ensure that the
            polygons are not too close together.
        anc : int, optional
            The number of ancilla qubits in the unit cell. Default is 1. If anc is set
            to 1, ancilla will be placed at the center of the polygon and at the center
            of the edges of the polygon.
        anc_radius : float | int, optional
            The radius of the circle on which the ancilla points are placed. Default is
            0.2.
        anc_offset : float | int, optional
            The offset to be applied to the angle of the ancilla points on the circle.
            This is useful for placing the first ancilla point at a specific angle.
            Default is -0.5 * np.pi (top of circle).
        anc_disp : tuple[int | float, int | float], optional
            The displacement of the ancilla points from the center of the circle. This
            is useful for placing the ancilla points at a specific position in the 2D
            plane. Default is (0, 0).

        Returns
        -------
        Lattice
            A Lattice object representing the polygonal lattice.
        """
        # Validation
        if n is None:
            raise ValueError("Please specify the number of sides `n` for the polygon.")
        if not isinstance(n, int) or n < 3:
            raise ValueError("Number of sides must be an integer that is at least 3.")
        if not isinstance(anc, int) or anc < 0:
            raise ValueError("Number of ancilla qubits must be a non-negative integer.")

        lattice_type = LatticeType.POLY_2D
        basis_vectors = cls._points_on_circle(n, poly_radius, poly_offset)
        if anc > 0:
            basis_vectors = basis_vectors + cls._points_on_circle(
                anc, anc_radius, anc_offset, anc_disp
            )
        cell_dist = poly_radius * cell_dist_factor
        lattice_vectors = [[0, cell_dist], [cell_dist, 0]]
        return cls(basis_vectors, lattice_vectors, lattice_size, lattice_type)

    @classmethod
    def cube_3d(
        cls,
        lattice_size: tuple[int, int, int] | None = None,
    ):
        """
        Generates a cubic lattice in 3D. The lattice is defined by three lattice
        vectors that are orthogonal to each other tiling the 3D space in a cubic
        pattern.

        Parameters
        ----------
        lattice_size : tuple[int, int, int] | None, optional
            The size of the lattice in terms of unit cells. If None, the lattice is
            assumed to be infinitely large.

        Returns
        -------
        Lattice
            A Lattice object representing the cubic lattice.
        """
        lattice_type = LatticeType.CUBE_3D
        basis_vectors = []
        lattice_vectors = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
        return cls(basis_vectors, lattice_vectors, lattice_size, lattice_type)



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from copy import deepcopy
from typing import Iterator, Tuple

from .moments.base_moment import Moment
from .operations.base_operation import Operation
from .operations.gate_operation import (
    GateOperation,
    TwoQubitGateOperation,
)
from .operations.resize_operation import ResizeOperation
from .operations.measurement_operation import MeasurementOperation
from .operations.datamanipulation_operation import (
    DataManipulationOperation,
)
from .operations.classical_operation import ClassicalOperation
from .operations.controlled_operation import ControlledOperation


class MomentQueue:
    """This object takes an input of a list of Operation objects, creates a
    Moment objects based on the list then manages the sequence of Moment
    objects.

    The Moment objects will then be sent to the Tableau and
    transform the internal representation.
    """

    def __init__(
        self, input_operations: list[Operation], parallelize: bool = False
    ) -> None:
        self.input_operations = deepcopy(input_operations)
        if parallelize:
            self.parallelized_operations = []
            self._parallel_operation_split(deepcopy(self.input_operations))
        else:
            self.parallelized_operations = [
                [op] for op in deepcopy(self.input_operations)
            ]

        self.moment_generators = self._create_generators(self.parallelized_operations)

    def _create_generators(
        self, parallelised_operations: list[Operation]
    ) -> Iterator[Tuple[Moment, Moment]]:
        """Create a generator that returns a tuple of Moment objects. The first
        Moment object is the forward moment and the second is the backward
        moment.
        """
        n_moments = len(parallelised_operations)
        for i in range(n_moments):
            yield (
                MomentFactory.create_moment(parallelised_operations[i], i),
                MomentFactory.create_moment(
                    parallelised_operations[-i - 1], n_moments - i - 1
                ),
            )

    def _parallel_operation_split(self, input_operations: list[Operation]) -> bool:
        """Parallelize GateOperation Objects in the input list, by returning a
        list of lists."""
        # To Create Exception for Operations that dont have target_qubit.
        while len(input_operations) != 0:
            first_op = input_operations.pop(0)
            parallel_op = [first_op]
            if isinstance(
                first_op,
                (DataManipulationOperation, ClassicalOperation, ControlledOperation),
            ):
                occupied_qubit = []
            else:
                occupied_qubit = [first_op.target_qubit]
            if isinstance(first_op, TwoQubitGateOperation):
                occupied_qubit.append(first_op.control_qubit)
            if isinstance(first_op, GateOperation):
                pop_index = []
                for op_index, next_op in enumerate(input_operations):
                    # Measurement, Resize and DataManipulation Operations are hard walls
                    # that prevent other Operations from being performed
                    if isinstance(
                        next_op,
                        (
                            MeasurementOperation,
                            ResizeOperation,
                            DataManipulationOperation,
                            ClassicalOperation,
                            ControlledOperation,
                        ),
                    ):
                        break
                    if isinstance(next_op, GateOperation):
                        if [
                            qubit_number
                            for qubit_number in next_op.operating_qubit
                            if qubit_number in occupied_qubit
                        ] == []:
                            pop_index.append(op_index)
                            parallel_op.append(next_op)
                        occupied_qubit.extend(next_op.operating_qubit)
                sort_index = sorted(pop_index, reverse=True)
                for each_index in sort_index:
                    input_operations.pop(each_index)
            self.parallelized_operations.append(parallel_op)
        return True

    @property
    def input_operations(self) -> list[Operation]:
        """
        The list of Operation objects that are to be converted into Moment
        objects.
        """
        return self._input_operations

    @input_operations.setter
    def input_operations(self, input_operations: list[Operation]) -> None:
        for each_item in input_operations:
            if not isinstance(each_item, Operation):
                raise TypeError(
                    "The input_operations should be a list of Operation objects."
                )
        self._input_operations = input_operations

    def reset_queue(self):
        """Resets the queue to contain all the instructions again."""
        # self.queue = self.queue_full.copy()
        self.moment_generators = self._create_generators(self.parallelized_operations)


class MomentFactory:  # pylint: disable=too-few-public-methods
    """This object is in charge of creating Moment Objects."""

    @staticmethod
    def create_moment(input_operations: list[Operation], time_step: int) -> Moment:
        """
        Create a Moment object from a list of Operation objects.
        """
        if not input_operations:
            raise ValueError(
                """input_operations cannot be empty. There should be at least 1
                Operation in 1 Moment."""
            )
        # Elements are True if all the inputs are valid Operation Types
        gate_check = [
            isinstance(
                each_operations,
                (
                    GateOperation,
                    MeasurementOperation,
                    ResizeOperation,
                    DataManipulationOperation,
                    ClassicalOperation,
                    ControlledOperation,
                ),
            )
            for each_operations in input_operations
        ]
        if all(gate_check):
            return Moment(input_operations, time_step)
        raise TypeError(
            "There are invalid Operation types in the input_operations list."
        )



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from dataclasses import dataclass, field

import numpy as np

from .base_operation import Operation, OpType
from ..pauli_frame import PauliFrame
from .controlled_operation import has_ccontrol


@dataclass
@has_ccontrol
class DataManipulationOperation(Operation):
    """
    Operations of this type manipulate data within the Engine during runtime.
    """

    operation_type: str = field(default=OpType.DATAMANIPULATION, init=False)


@dataclass
class UpdateTableau(DataManipulationOperation):
    """
    An Operation that updates the state of the Tableau in the Engine during runtime.
    Note that the tableau must be a numpy array with bits, 0s and 1s, for every element.

    Parameters
    ----------
    tableau: numpy.ndarray
        The state of the Tableau to be updated to.
    validate: bool
        The tableau will be validated at runtime. The tableau provided must be a valid
        stabilizer tableau. The default is True.
    """

    name: str = field(default="UpdateTableau", init=False)
    tableau: np.ndarray
    validate: bool = field(default=True)
    # Maybe add tableau valdiation here instead? Currently done in Instructions if
    # validate = True.


@dataclass
class CreatePauliFrame(DataManipulationOperation):
    """
    An Operation that creates a Pauli Frame within the Engine during runtime. The
    Pauli Frame created will be associated with the name provided by the user, and can
    be accessed at runtime by other operations. Since there can be multiple Pauli
    Frames, the frames will be associated by their names.
    """

    name: str = field(default="CreatePauliFrame", init=False)
    pauli_frame: PauliFrame

    def __post_init__(self):
        # Check validity of pauli_frame
        if not isinstance(self.pauli_frame, PauliFrame):
            raise TypeError(
                f"Invalid PauliFrame '{self.pauli_frame}'. Must be of type "
                "'PauliFrame'."
            )


@dataclass
class RecordPauliFrame(DataManipulationOperation):
    """
    Records the current state of the Pauli Frame into the DataStore. The state can then
    be accessed at the end of the run from the DataStore. The Pauli Frame to be recorded
    will be identified by its name.
    """

    name: str = field(default="RecordPauliFrame", init=False)
    pauli_frame: PauliFrame

    def __post_init__(self):
        # Check validity of pauli_frame
        if not isinstance(self.pauli_frame, PauliFrame):
            raise TypeError(
                f"Invalid PauliFrame '{self.pauli_frame}'. Must be of type "
                "'PauliFrame'."
            )


@dataclass
class CreateClassicalRegister(DataManipulationOperation):
    """
    An operation that creates a Classical Register within the Engine that can be
    accessed at runtime by classical operations. Since there can be multiple classical
    registers, the registers will be associated by their reg_name, and their bits by
    their respective bit IDs or bit ordering as provided by the user.

    If trying to initialize a classical register with bit IDs, the number of bit IDs
    must be equal to the number of bits in the register.

    Example of valid Classical Register:
    CreateClassicalRegister(reg_name="testreg", no_of_bits=3, bit_ids=["bit_id_1",
    "bit_id_2", "bit_id_3"])

    Parameters
    ----------
    reg_name: str
        The name of the classical register. This name will be used by cliffordsim to
        identify the classical register being referred to by the user.
    no_of_bits: int
        The number of bits to initialize the register with. Default is 1.
    bit_ids: list[str]
        The bit IDs of all the bits in the register. Example input: ["bit_id_1",
        "bit_id_2", "bit_id_3"].
        The number of bit IDs provided must be equal to the number of bits in the
        classical register.
        If no bit IDs are provided, the classical register will randomly generate
        the IDs upon initialization.
    """

    name: str = field(default="CreateClassicalRegister", init=False)
    reg_name: str
    no_of_bits: int = field(default=1)
    bit_ids: list[str] = field(default=None)

    def __post_init__(self):
        # Check validity of bit IDs
        if self.bit_ids is not None:
            for each_id in self.bit_ids:
                if not isinstance(each_id, str):
                    raise TypeError("The bit ID must be a string.")
            if len(self.bit_ids) != self.no_of_bits:
                raise ValueError(
                    "The number of bit IDs must be equal to the number of bits in the "
                    "register."
                )


@dataclass
class RecordClassicalRegister(DataManipulationOperation):
    """
    Records the current state of the Classical Register into the DataStore. The state
    can then be accessed at the end of the run from the DataStore.

    Parameters
    ----------
    reg_name: str
        The name of the Classical Register to record.
    """

    name: str = field(default="RecordClassicalRegister", init=False)
    reg_name: str
